# -*- coding: utf-8 -*-
"""
Created on Thu Aug 27 16:39:32 2015

@author: ebachelet
"""
from __future__ import division
import time as python_time

import numpy as np
import scipy.optimize
import collections
import warnings
import emcee
import sys
import copy
from collections import OrderedDict


from pyLIMA import microlmodels
from pyLIMA import microloutputs
from pyLIMA import microlguess
from pyLIMA import microltoolbox
from pyLIMA import microlpriors
from pyLIMA import microlcaustics

warnings.filterwarnings("ignore")


class MLFits(object):
    """
    ######## Fitter module ########

    This class contains the method to fit the event with the selected attributes.

    **WARNING**: All fits (and so results) are made using data in flux.

    Attributes :

        event : the event object on which you perform the fit on. More details on the event module.

        model : The microlensing model you want to fit. Has to be an object define in
                microlmodels module.
                More details on the microlmodels module.

        method : The fitting method you want to use for the fit.

        guess : The guess you can give to the fit or the guess return by the initial_guess function.

        fit_results : the fit parameters returned by method LM and DE.

        fit_covariance : the fit parameters covariance matrix returned by method LM and DE.

        fit_time : the time needed to fit.

        MCMC_chains : the MCMC chains returns by the MCMC method

        MCMC_probabilities : the objective function computed for each chains of the MCMC method

        fluxes_MCMC_method : a string describing how you want to estimate the model fluxes for the MCMC method.

        outputs : the standard pyLIMA outputs. More details in the microloutputs module.

    :param object event: the event object on which you perform the fit on. More details on the
                         event module.


    """

    def __init__(self, event):
        """The fit class has to be intialized with an event object."""

        self.event = event
        self.model = microlmodels.ModelPSPL(event)
        self.method = 'None'
        self.guess = []
        self.outputs = []
        self.fit_results = []
        self.fit_covariance = []
        self.fit_time = []
        self.DE_population = []
        self.binary_regime = None
        self.MCMC_chains = []
        self.MCMC_probabilities = []
        self.fluxes_MCMC_method = ''
        self.pool = None

    def mlfit(self, model, method, DE_population_size=10, flux_estimation_MCMC='MCMC', fix_parameters_dictionnary=None,
              grid_resolution=10, computational_pool=None, binary_regime=None):
        """This function realize the requested microlensing fit, and set the according results
        attributes.

        :param object model: the model object requested. More details on the microlmodels module.

        :param string method: The fitting method you want to use. Has to be a string  in :

                                 'LM' --> Levenberg-Marquardt algorithm. Based on the
                                 scipy.optimize.leastsq routine.
                                          **WARNING** : the parameter maxfev (number of maximum
                                          iterations) is set to 50000
                                          the parameter ftol (relative precision on the chi^2) is
                                          set to 0.00001
                                          your fit may not converge because of these limits.
                                          The starting points of this method are found using the
                                          initial_guess method.
                                          Obviously, this can fail. In this case, switch to
                                          method 'DE'.

                                 'DE' --> Differential evolution algoritm. Based on the
                                 scipy.optimize.differential_evolution.
                                          Look Storn & Price (1997) : "Differential Evolution – A
                                          Simple and Efficient Heuristic for global Optimization
                                          over Continuous Spaces"
                                          Because this method is heuristic, it is not 100% sure a
                                          satisfying solution is found. Just relaunch :)
                                          The result is then use as a starting point for the 'LM'
                                          method.


                                 'MCMC' --> Monte-Carlo Markov Chain algorithm. Based on the
                                 emcee python package :
                                          " emcee: The MCMC Hammer" (Foreman-Mackey et al. 2013).
                                          The inital population is computed around the best
                                          solution returned by
                                          the 'DE' method.


        :param int DE_population_size:  The population factor desired for the DE method. Default is 10.

        :param string flux_estimation_MCMC: The desired method to estimate the fluxes (f_source and g) of the
                                             telescopes. 'MCMC' will do this through an MCMC method (default) when
                                             everything else will do this thanks to a 1D polyfit through np.polyfit.

        Note that a sanity check is done post-fit to assess the fit quality with the check_fit
        function.
        """
        print('')
        print('Start fit on ' + self.event.name + ', with model ' + model.model_type + ' and method ' + method)
        self.event.check_event()

        self.model = model
        self.method = method
        self.fluxes_MCMC_method = flux_estimation_MCMC
        self.DE_population_size = DE_population_size
        self.model.define_model_parameters()
        if computational_pool:
            from mpi4py import MPI
            import dill

            MPI.pickle.dumps = dill.dumps
            MPI.pickle.loads = dill.loads
            self.pool = computational_pool
        self.binary_regime = binary_regime

        if self.method == 'LM':
            number_of_data = self.event.total_number_of_data_points()
            if number_of_data <= (len(self.model.model_dictionnary)):

                print("You do not have enough data points to use this method (LM), please switch to other methods." \
                      " Given the requested total model " + str(self.model.model_dictionnary.keys()) + \
                      " you need at least " + str(
                    len(self.model.model_dictionnary)) + ' data points to use the method LM!')
                return

            else:

                self.fit_results, self.fit_covariance, self.fit_time = self.lmarquardt()

        if self.method == 'TRF':
            self.fit_results, self.fit_covariance, self.fit_time = self.trust_region_reflective()
        if self.method == 'DE':
            self.fit_results, self.fit_covariance, self.fit_time = self.differential_evolution()

        if self.method == 'MCMC':
            self.MCMC_chains = self.MCMC()

        if self.method == 'GRIDS':
            self.fix_parameters_dictionnary = OrderedDict(
                sorted(fix_parameters_dictionnary.items(), key=lambda x: x[1]))
            self.grid_resolution = grid_resolution
            self.grids()

        fit_quality_flag = 'Good Fit'

        if self.method != 'MCMC':
            fit_quality_flag = self.check_fit()

        if fit_quality_flag == 'Bad Fit':

            if self.method == 'LM':

                print('We have to change method, this fit was unsuccessfull. We decided to switch ' \
                      '' \
                      'method to "DE"')

                # self.method = 'DE'
                # self.mlfit(self.model, self.method, self.fluxes_MCMC_method)

            else:

                print('Unfortunately, this is too hard for pyLIMA :(')

    def check_fit(self):
        """Check if the fit results and covariance make sens.

         0.0 terms or a negative term in the diagonal covariance matrix indicate the fit is not
         reliable.

         A negative source flux is also counted as a bad fit.

         A negative rho or rho> 0.1 is also consider as a bad fit

         :return: a flag indicated good or bad fit ('Good Fit' or 'Bad Fit')
         :rtype: string
        """

        flag_quality = 'Good Fit'
        negative_covariance_diagonal = np.diag(self.fit_covariance) < 0
        number_of_data = self.event.total_number_of_data_points()
        if number_of_data >= (len(self.model.model_dictionnary) + 2 * len(self.event.telescopes)):

            if (0.0 in self.fit_covariance):
                print('Your fit probably wrong. Cause ==> bad covariance matrix')
                flag_quality = 'Bad Fit'
                return flag_quality

        if (True in negative_covariance_diagonal) | \
                (np.isnan(self.fit_covariance).any()) | (np.isinf(self.fit_covariance).any()):
            print('Your fit probably wrong. Cause ==> bad covariance matrix')
            flag_quality = 'Bad Fit'
            return flag_quality

        for i in self.event.telescopes:

            if self.fit_results[self.model.model_dictionnary['fs_' + i.name]] < 0:
                print('Your fit probably wrong. Cause ==> negative source flux for telescope ' + \
                      i.name)
                flag_quality = 'Bad Fit'
                return flag_quality

        if 'rho' in list(self.model.model_dictionnary.keys()):

            if (self.fit_results[self.model.model_dictionnary['rho']] > 0.1) | \
                    (self.fit_results[self.model.model_dictionnary['rho']] < 0.0):
                print('Your fit probably wrong. Cause ==> bad rho ')
                flag_quality = 'Bad Fit'
                return flag_quality

        return flag_quality

    def initial_guess(self):
        """Try to estimate the microlensing parameters. Only use for PSPL and FSPL
           models. More details on microlguess module.

           :return guess_parameters: a list containing parameters guess related to the model.
           :rtype: list
        """

        if len(self.model.parameters_guess) == 0:

            # Estimate  the Paczynski parameters

            if self.model.model_type == 'PSPL':
                guess_paczynski_parameters, f_source = microlguess.initial_guess_PSPL(self.event)

            if self.model.model_type == 'FSPL':
                guess_paczynski_parameters, f_source = microlguess.initial_guess_FSPL(self.event)

            if self.model.model_type == 'DSPL':
                guess_paczynski_parameters, f_source = microlguess.initial_guess_DSPL(self.event)

            # Estimate  the telescopes fluxes (flux_source + g_blending) parameters, with a PSPL model

            fake_model = microlmodels.create_model('PSPL', self.event)
            fake_model.define_model_parameters()
            telescopes_fluxes = self.find_fluxes(guess_paczynski_parameters, fake_model)

            # The survey fluxes are already known from microlguess
            telescopes_fluxes[0] = f_source
            telescopes_fluxes[1] = 0.0

            if self.model.parallax_model[0] != 'None':
                guess_paczynski_parameters = guess_paczynski_parameters + [0.0, 0.0]

            if self.model.xallarap_model[0] != 'None':
                guess_paczynski_parameters = guess_paczynski_parameters + [0, 0]

            if self.model.orbital_motion_model[0] != 'None':
                guess_paczynski_parameters = guess_paczynski_parameters + [0, 0]

            if self.model.source_spots_model != 'None':
                guess_paczynski_parameters = guess_paczynski_parameters + [0]



        else:

            guess_paczynski_parameters = list(self.model.parameters_guess)

            telescopes_fluxes = self.find_fluxes(guess_paczynski_parameters, self.model)

        guess_paczynski_parameters += telescopes_fluxes

        guess_parameters_pyLIMA_standards = collections.namedtuple('parameters',
                                                                   self.model.pyLIMA_standards_dictionnary.keys())

        for key_parameter in self.model.pyLIMA_standards_dictionnary.keys():

            try:
                setattr(guess_parameters_pyLIMA_standards, key_parameter,
                        guess_paczynski_parameters[self.model.pyLIMA_standards_dictionnary[key_parameter]])

            except:

                pass

        fancy_parameters_guess = self.model.pyLIMA_standard_parameters_to_fancy_parameters(
            guess_parameters_pyLIMA_standards)

        model_guess_parameters = []
        for key_parameter in list(self.model.model_dictionnary.keys()):
            model_guess_parameters.append(getattr(fancy_parameters_guess, key_parameter))

        print(sys._getframe().f_code.co_name, ' : Initial parameters guess SUCCESS')
        return model_guess_parameters

    def MCMC(self):
        """ The MCMC method. Construct starting points of the chains around
            the best solution found by the 'DE' method.
            The objective function is :func:`chichi_MCMC`. Telescope flux (fs and g), can be optimized thanks to MCMC if
            flux_estimation_MCMC is 'MCMC', either they are derived through np.polyfit.

            Based on the emcee python package :
            " emcee: The MCMC Hammer" (Foreman-Mackey et al. 2013).
            Have a look here : http://dan.iel.fm/emcee/current/

            :return: a tuple containing (MCMC_chains, MCMC_probabilities)
            :rtype: tuple

            **WARNING** :
                   nwalkers is set to 100
                   nlinks is set to 300
                   5*nwalkers*nlinks MCMC steps in total
        """

        nwalkers = 300
        nlinks = 100

        # start = python_time.time()

        if len(self.model.parameters_guess) == 0:

            differential_evolution_estimation = self.differential_evolution()[0]
            self.DE_population_size = 10
            self.guess = differential_evolution_estimation

        else:

            self.guess = list(self.model.parameters_guess)
            self.guess += self.find_fluxes(self.guess, self.model)

        # Best solution

        best_solution = self.guess

        limit_parameters = len(self.model.parameters_boundaries)

        # Initialize the population of MCMC
        population = []

        count_walkers = 0

        while count_walkers < nwalkers:

            # Construct an individual of the population around the best solution.
            individual = []
            for parameter_key in list(self.model.model_dictionnary.keys())[:limit_parameters]:

                parameter_trial = microlguess.MCMC_parameters_initialization(parameter_key,
                                                                             self.model.model_dictionnary,
                                                                             best_solution)

                if parameter_trial:

                    for parameter in parameter_trial:
                        individual.append(parameter)

            if self.fluxes_MCMC_method == 'MCMC':
                fluxes = self.find_fluxes(individual, self.model)
                individual += fluxes

            chichi = self.chichi_MCMC(individual)

            if chichi != -np.inf:
                # np.array(individual)
                # print count_walkers

                population.append(np.array(individual))
                count_walkers += 1

        print('pre MCMC done')

        number_of_parameters = len(individual)

        # pool = MPIPool()
        # if not pool.is_master():
        # pool.wait()
        # sys.exit(0)
        sampler = emcee.EnsembleSampler(nwalkers, number_of_parameters, self.chichi_MCMC,
                                        a=2.0, pool=self.pool)

        # First estimation using population as a starting points.

        final_positions, final_probabilities, state = sampler.run_mcmc(population, nlinks)

        print('MCMC preburn done')

        sampler.reset()
        MCMC_chains = None
        import pdb;
        pdb.set_trace()
        # Final estimation using the previous output.
        for positions, probabilities, states in sampler.sample(final_positions, iterations=5 * nlinks,
                                                               storechain=True):
            chains = np.c_[positions, probabilities]
            if MCMC_chains is not None:

                MCMC_chains = np.r_[MCMC_chains, chains]
            else:

                MCMC_chains = chains

        print(sys._getframe().f_code.co_name, ' : MCMC fit SUCCESS')
        return MCMC_chains

    def chichi_MCMC(self, fit_process_parameters):
        """Return the chi^2 for the MCMC method. There is some priors here.

        :param list fit_process_parameters: the model parameters ingested by the correpsonding
                                            fitting routine.

        :returns: here, the return is -chi^2/2 (likelihood)

        :rtype: float
        """

        chichi = 0

        pyLIMA_parameters = self.model.compute_pyLIMA_parameters(fit_process_parameters)

        for telescope in self.event.telescopes:
            # Find the residuals of telescope observation regarding the parameters and model
            residus = self.model_residuals(telescope, pyLIMA_parameters)

            chichi += (residus ** 2).sum()

        return -chichi / 2

    def differential_evolution(self):
        """  The DE method. Differential evolution algorithm. The objective function is
        :func:`chichi_differential_evolution`. The flux parameters are estimated through np.polyfit.
         Based on the scipy.optimize.differential_evolution.
         Look Storn & Price (1997) :
         "Differential Evolution – A Simple and Efficient Heuristic for
         global Optimization over Continuous Spaces"

         :return: a tuple containing (fit_results, fit_covariance, computation_time)
         :rtype: tuple

         **WARNING** :
                   tol (relative standard deviation of the objective function) is set to 10^-4

                   popsize (the total number of individuals is :
                   popsize*number_of_paczynski_parameters)
                   is set to DE_population_size

                   mutation is set to (0.1, 1.5)

                   recombination is set to 0.7

                   These parameters can avoid the fit to properly converge (expected to be rare :)).
                   Just relaunch should be fine.
        """

        starting_time = python_time.time()
        differential_evolution_estimation = scipy.optimize.differential_evolution(
            self.chichi_differential_evolution,
            bounds=self.model.parameters_boundaries,
            mutation=(0.5, 1.5), popsize=int(self.DE_population_size), maxiter=5000, tol=0.0,
            atol=0.001, strategy='best1bin',
            recombination=0.5, polish=True, init='latinhypercube',
            disp=True
        )

        # paczynski_parameters are all parameters to compute the model, excepted the telescopes fluxes.
        paczynski_parameters = differential_evolution_estimation['x'].tolist()

        print('DE converge to objective function : f(x) = ', str(differential_evolution_estimation['fun']))
        print('DE converge to parameters : = ', differential_evolution_estimation['x'].astype(str))

        # Construct the guess for the LM method. In principle, guess and outputs of the LM
        # method should be very close.

        number_of_data = self.event.total_number_of_data_points()
        if number_of_data <= (len(self.model.model_dictionnary)):

            print("You do not have enough data points to use LM method to estimate the covariance matrix." \
                  "The covariance matrix is set to 0.0. please switch to MCMC if you need errors estimation.")

            fit_results = paczynski_parameters + self.find_fluxes(paczynski_parameters, self.model) + \
                          [differential_evolution_estimation['fun']]
            fit_covariance = np.zeros((len(paczynski_parameters) + 2 * len(self.event.telescopes),
                                       len(paczynski_parameters) + 2 * len(self.event.telescopes)))

        else:

            self.guess = paczynski_parameters + self.find_fluxes(paczynski_parameters, self.model)

            fit_results, fit_covariance, fit_time = self.lmarquardt()

        computation_time = python_time.time() - starting_time

        print(sys._getframe().f_code.co_name, ' : Differential evolution fit SUCCESS')
        return fit_results, fit_covariance, computation_time

    def chichi_differential_evolution(self, fit_process_parameters):
        """Return the chi^2 for the DE method.

        :param list fit_process_parameters: the model parameters ingested by the correpsonding
        fitting routine.

        :returns: the chi^2

        :rtype: float
        """
        pyLIMA_parameters = self.model.compute_pyLIMA_parameters(fit_process_parameters)

        try:
            self.model.x_center = None
            self.model.y_center = None
        except:
            pass
        chichi = 0.0

        for telescope in self.event.telescopes:
            # Find the residuals of telescope observation regarding the parameters and model
            residus = self.model_residuals(telescope, pyLIMA_parameters)

            chichi += (residus ** 2).sum()

        self.DE_population.append(fit_process_parameters.tolist() + [chichi])

        return chichi

    def lmarquardt(self):
        """The LM method. This is based on the Levenberg-Marquardt algorithm:

           "A Method for the Solution of Certain Problems in Least Squares"
           Levenberg, K. Quart. Appl. Math. 2, 1944, p. 164-168
           "An Algorithm for Least-Squares Estimation of Nonlinear Parameters"
           Marquardt, D. SIAM J. Appl. Math. 11, 1963, p. 431-441

           Based on scipy.optimize.leastsq python routine, which is based on MINPACK's lmdif and
           lmder
           algorithms (fortran based).

           The objective function is :func:`residuals_LM`.
           The starting point parameters are self.guess.
           the Jacobian is given by :func:`LM_Jacobian`.

           The fit is performed on all parameters : Paczynski parameters and telescopes fluxes.

           :return: a tuple containing (fit_results, covariance_matrix, computation_time)
           :rtype: tuple

           **WARNING**:
                     ftol (relative error desired in the sum of square) is set to 10^-6
                     maxfev (maximum number of function call) is set to 50000
                     These limits can avoid the fit to properly converge (expected to be rare :))
        """
        starting_time = python_time.time()

        # use the analytical Jacobian (faster) if no second order are present, else let the
        # algorithm find it.
        if self.guess == []:
            self.guess = self.initial_guess()
        n_data = 0
        for telescope in self.event.telescopes:
            n_data = n_data + telescope.n_data('flux')

        n_parameters = len(self.model.model_dictionnary)

        if self.model.Jacobian_flag == 'OK':
            lmarquardt_fit = scipy.optimize.leastsq(self.residuals_LM, self.guess, maxfev=50000,
                                                    Dfun=self.LM_Jacobian, col_deriv=0, full_output=1, ftol=10 ** -8,
                                                    xtol=10 ** -10, gtol=10 ** -10)
            fit_result = lmarquardt_fit[0].tolist()
            fit_result.append(microltoolbox.chichi(self.residuals_LM, lmarquardt_fit[0]))

            try:
            # Try to extract the covariance matrix from the lmarquard_fit output

                covariance_matrix = lmarquardt_fit[1] * fit_result[-1] / (n_data - n_parameters)

            except:

                covariance_matrix = np.zeros((len(self.model.model_dictionnary),
                                              len(self.model.model_dictionnary)))
        else:


            lmarquardt_fit = scipy.optimize.least_squares(self.residuals_LM, self.guess, method='lm', x_scale='jac', ftol=10 ** -10,
                                                          xtol=10 ** -10, gtol=10 ** -10,
                                                         )
            fit_result = lmarquardt_fit['x'].tolist()
            fit_result.append(microltoolbox.chichi(self.residuals_LM, lmarquardt_fit['x']))

            try:
                # Try to extract the covariance matrix from the lmarquard_fit output
                jacobian = lmarquardt_fit['jac']

                covariance_matrix = np.linalg.inv(np.dot(jacobian.T, jacobian))



            except:

                covariance_matrix = np.zeros((len(self.model.model_dictionnary),
                                              len(self.model.model_dictionnary)))


        computation_time = python_time.time() - starting_time


        # import pdb; pdb.set_trace()
        print(sys._getframe().f_code.co_name, ' : Levenberg_marquardt fit SUCCESS')
        print(fit_result)
        return fit_result, covariance_matrix, computation_time


    def residuals_LM(self, fit_process_parameters):
        """The normalized residuals associated to the model and parameters.

           :param list fit_process_parameters: the model parameters ingested by the correpsonding
                                               fitting routine.

           :return: a numpy array which represents the residuals_i for each telescope,
                    residuals_i=(data_i-model_i)/sigma_i
           :rtype: array_like
           The sum of square residuals gives chi^2.
        """

        pyLIMA_parameters = self.model.compute_pyLIMA_parameters(fit_process_parameters)
        try:
            self.model.x_center = None
            self.model.y_center = None
        except:
            pass
        residuals = np.array([])

        for telescope in self.event.telescopes:
            # Find the residuals of telescope observation regarding the parameters and model
            residus = self.model_residuals(telescope, pyLIMA_parameters)

            residuals = np.append(residuals, residus)
        # print python_time.time()-start

        return residuals

    def LM_Jacobian(self, fit_process_parameters):
        """Return the analytical Jacobian matrix, if requested by method LM.
        Available only for PSPL and FSPL without second_order.

        :param list fit_process_parameters: the model parameters ingested by the correpsonding
                                            fitting routine.
        :return: a numpy array which represents the jacobian matrix
        :rtype: array_like
        """

        pyLIMA_parameters = self.model.compute_pyLIMA_parameters(fit_process_parameters)

        count = 0
        # import pdb;
        # pdb.set_trace()
        for telescope in self.event.telescopes:

            if count == 0:

                _jacobi = self.model.model_Jacobian(telescope, pyLIMA_parameters)

            else:

                _jacobi = np.c_[_jacobi, self.model.model_Jacobian(telescope, pyLIMA_parameters)]

            count += 1

        # The objective function is : (data-model)/errors

        _jacobi = -_jacobi
        jacobi = _jacobi[:-2]
        # Split the fs and g derivatives in several columns correpsonding to
        # each observatories
        start_index = 0
        dresdfs = _jacobi[-2]
        dresdg = _jacobi[-1]

        for telescope in self.event.telescopes:
            derivative_fs = np.zeros((len(dresdfs)))
            derivative_g = np.zeros((len(dresdg)))
            index = np.arange(start_index, start_index + len(telescope.lightcurve_flux[:, 0]))
            derivative_fs[index] = dresdfs[index]
            derivative_g[index] = dresdg[index]
            jacobi = np.r_[jacobi, np.array([derivative_fs, derivative_g])]

            start_index = index[-1] + 1

        return jacobi.T

    def trust_region_reflective(self):

        starting_time = python_time.time()

        # use the analytical Jacobian (faster) if no second order are present, else let the
        # algorithm find it.
        if self.guess == []:
            self.guess = self.initial_guess()

        bounds_min = [i[0] for i in self.model.parameters_boundaries] + [0, -np.inf] * len(self.event.telescopes)
        bounds_max = [i[1] for i in self.model.parameters_boundaries] + [np.inf, np.inf] * len(self.event.telescopes)

        if self.model.Jacobian_flag == 'OK':
            trf_fit = scipy.optimize.least_squares(self.residuals_LM, self.guess, max_nfev=50000,
                                                   jac=self.LM_Jacobian, bounds=(bounds_min, bounds_max), ftol=10 ** -6,
                                                   xtol=10 ** -10, gtol=10 ** -5)
        else:

            trf_fit = scipy.optimize.least_squares(self.residuals_LM, self.guess, max_nfev=50000,
                                                   bounds=(bounds_min, bounds_max), ftol=10 ** -6,
                                                   xtol=10 ** -10, gtol=10 ** -5)
        computation_time = python_time.time() - starting_time

        fit_result = np.copy(trf_fit['x']).tolist()
        fit_result += [2 * trf_fit['cost']]

        try:

            jacobian = trf_fit['jac']

        except:

            jacobian = self.LM_Jacobian(fit_result)

        covariance_matrix = np.linalg.inv(np.dot(jacobian.T, jacobian))
        n_data = 0
        for telescope in self.event.telescopes:
            n_data = n_data + telescope.n_data('flux')

        n_parameters = len(self.model.model_dictionnary)
        covariance_matrix *= fit_result[-1] / (n_data - n_parameters)
        print(sys._getframe().f_code.co_name, ' : TRF fit SUCCESS')
        print(fit_result)
        return fit_result, covariance_matrix, computation_time

    def chichi_telescopes(self, fit_process_parameters):
        """Return a list of chi^2 (float) for individuals telescopes.

        :param list fit_process_parameters: the model parameters ingested by the correpsonding
                                            fitting routine.

        :returns: the chi^2 for each telescopes

        :rtype: list
        """

        residuals = self.residuals_LM(fit_process_parameters)
        chichi_list = []
        start_index = 0
        for telescope in self.event.telescopes:
            chichi_list.append(
                (residuals[start_index:start_index + len(telescope.lightcurve_flux)] ** 2).sum())

            start_index += len(telescope.lightcurve_flux)

        return chichi_list

    def model_residuals(self, telescope, pyLIMA_parameters):
        """ Compute the residuals of a telescope lightcurve according to the model.

        :param object telescope: a telescope object. More details in telescopes module.
        :param object pyLIMA_parameters: object containing the model parameters, see microlmodels for more details

        :return: the residuals in flux, the priors
        :rtype: array_like, float
        """
        lightcurve = telescope.lightcurve_flux

        flux = lightcurve[:, 1]
        errflux = lightcurve[:, 2]

        microlensing_model = self.model.compute_the_microlensing_model(telescope, pyLIMA_parameters)

        residuals = (flux - microlensing_model[0]) / errflux

        return residuals

    def all_telescope_residuals(self, pyLIMA_parameters):
        """ Compute the residuals of all telescopes according to the model.

        :param object pyLIMA_parameters: object containing the model parameters, see microlmodels for more details

        :return: the residuals in flux,
        :rtype: list, a list of array of residuals in flux
        """

        residuals = []
        for telescope in self.event.telescopes:
            # Find the residuals of telescope observation regarding the parameters and model
            residus = self.model_residuals(telescope, pyLIMA_parameters)
            # no prior here
            residuals.append(residus)
        # print python_time.time()-start
        return residuals

    def find_fluxes(self, fit_process_parameters, model):
        """Find telescopes flux associated (fs,g) to the model. Used for initial_guess and LM
        method.

        :param list fit_process_parameters: the model parameters ingested by the correpsonding fitting
                                       routine.
        :param object model: a microlmodels which you want to compute the fs,g parameters.

        :return: a list of tuple with the (fs,g) telescopes flux parameters.
        :rtype: list
        """

        telescopes_fluxes = []
        pyLIMA_parameters = model.compute_pyLIMA_parameters(fit_process_parameters)

        for telescope in self.event.telescopes:

            flux = telescope.lightcurve_flux[:, 1]

            ml_model, f_source, f_blending = model.compute_the_microlensing_model(telescope, pyLIMA_parameters)

            # Prior here
            if f_source < 0:

                telescopes_fluxes.append(np.min(flux))
                telescopes_fluxes.append(0.0)
            else:
                telescopes_fluxes.append(f_source)
                telescopes_fluxes.append(f_blending)
        return telescopes_fluxes

    def grids(self):
        """ Compute models on a grid. ON CONSTRUCTION.
        """

        parameters_on_the_grid = []

        for parameter_name in self.fix_parameters_dictionnary:
            parameter_range = self.model.parameters_boundaries[self.model.model_dictionnary[parameter_name]]

            parameters_on_the_grid.append(
                np.linspace(parameter_range[0], parameter_range[1],
                            self.grid_resolution))

        hyper_grid = self.construct_the_hyper_grid(parameters_on_the_grid)

        self.new_parameters_boundaries = self.redefine_parameters_boundaries()

        if self.pool is not None:
            computational_map = self.pool.map

        else:
            computational_map = map

        grid_results = list(computational_map(self.optimization_on_grid_pixel, hyper_grid))
        import pdb;
        pdb.set_trace()
        return grid_results

    def optimization_on_grid_pixel(self, grid_pixel_parameters):

        differential_evolution_estimation = scipy.optimize.differential_evolution(
            self.chichi_grids,
            bounds=self.new_parameters_boundaries, args=tuple(grid_pixel_parameters.tolist()),
            mutation=(0.5, 1.0), popsize=1, maxiter=100,
            tol=0.0, atol=0.1, strategy='best1bin',
            recombination=0.7, polish=True, init='latinhypercube',
            disp=True
        )

        best_parameters = self.reconstruct_fit_process_parameters(differential_evolution_estimation['x'],
                                                                  grid_pixel_parameters)

        best_parameters += [differential_evolution_estimation['fun']]

        print(sys._getframe().f_code.co_name, ' Grid step on ' + str(grid_pixel_parameters.tolist()).strip(
            '[]') + ' converge to f(x) = ' + str(differential_evolution_estimation['fun']))

        return best_parameters

    def chichi_grids(self, moving_parameters, *fix_parameters):
        """ Compute chi^2. ON CONSTRUCTION.
        """
        fit_process_parameters = self.reconstruct_fit_process_parameters(moving_parameters, fix_parameters)

        pyLIMA_parameters = self.model.compute_pyLIMA_parameters(fit_process_parameters)

        chichi = 0.0
        for telescope in self.event.telescopes:
            # Find the residuals of telescope observation regarding the parameters and model

            residus = self.model_residuals(telescope, pyLIMA_parameters)

            chichi += (residus ** 2).sum()

        return chichi

    def reconstruct_fit_process_parameters(self, moving_parameters, fix_parameters):
        """ Reconstruc parameters. ON CONSTRUCTION.
        """
        fit_process_parameters = []

        for key in list(self.model.model_dictionnary.keys())[:len(self.model.parameters_boundaries)]:

            if key in self.moving_parameters_dictionnary:

                fit_process_parameters.append(moving_parameters[self.moving_parameters_dictionnary[key]])

            else:

                fit_process_parameters.append(fix_parameters[self.fix_parameters_dictionnary[key]])

        return fit_process_parameters

    def redefine_parameters_boundaries(self):
        """ Recompute the parameters boundaries. ON CONSTRUCTION.
        """
        parameters_boundaries = []
        self.moving_parameters_dictionnary = {}
        count = 0

        for indice, key in enumerate(list(self.model.model_dictionnary.keys())[:len(self.model.parameters_boundaries)]):

            if key not in self.fix_parameters_dictionnary.keys():
                parameters_boundaries.append(self.model.parameters_boundaries[indice])
                self.moving_parameters_dictionnary[key] = count
                count += 1
        return parameters_boundaries

    def construct_the_hyper_grid(self, parameters):
        """Define the grid. ON CONSTRUCTION.
        """
        params = map(np.asarray, parameters)
        grid = np.broadcast_arrays(*[x[(slice(None),) + (None,) * i] for i, x in enumerate(params)])

        reformate_grid = np.vstack(grid).reshape(len(parameters), -1).T
        return reformate_grid

    def produce_outputs(self):
        """ Produce the standard outputs for a fit.
        More details in microloutputs module.
        """

        if self.method != 'MCMC':

            outputs = microloutputs.LM_outputs(self)

        else:

            outputs = microloutputs.MCMC_outputs(self)

        self.outputs = outputs

    def produce_fit_statistics(self):
        """ Produce the standard outputs for a fit.
        More details in microloutputs module.
        """

        stats_outputs = microloutputs.statistical_outputs(self)

        self.stats_outputs = stats_outputs

    def produce_pdf(self, output_directory):
        """ ON CONSTRUCTION
        """
        microloutputs.pdf_output(self, output_directory)



# -*- coding: utf-8 -*-
"""
Created on Mon May 23 16:00:51 2016

@author: ebachelet
"""

import numpy as np
import scipy.signal as ss

from pyLIMA import microltoolbox


def initial_guess_PSPL(event):
    """Function to find initial PSPL guess for Levenberg-Marquardt solver (method=='LM').
       This assumes no blending.

    :param object event: the event object on which you perform the fit on. More details on the
                            event module.

    :return: the PSPL guess for this event. A list of parameters associated to the PSPL model + the source flux of
    :return: the PSPL guess for this event. A list of parameters associated to the PSPL model + the source flux of
                the survey telescope.
    :rtype: list,float
    """

    # to estimation
    to_estimations = []
    maximum_flux_estimations = []
    errors_magnitude = []

    for telescope in event.telescopes:
        # Lot of process here, if one fails, just skip
        lightcurve_magnitude = telescope.lightcurve_magnitude
        mean_error_magnitude = np.mean(lightcurve_magnitude[:, 2])
        try:

            # only the best photometry
            good_photometry_indexes = np.where((lightcurve_magnitude[:, 2] <
                                                max(0.1, mean_error_magnitude)))[0]
            lightcurve_bis = lightcurve_magnitude[good_photometry_indexes]

            lightcurve_bis = lightcurve_bis[lightcurve_bis[:, 0].argsort(), :]

            mag = lightcurve_bis[:, 1]
            flux = microltoolbox.magnitude_to_flux(mag)

            # clean the lightcurve using Savitzky-Golay filter on 3 points, degree 1.
            mag_clean = ss.savgol_filter(mag, 3, 1)
            time = lightcurve_bis[:, 0]
            flux_clean = microltoolbox.flux_to_magnitude(mag_clean)
            errmag = lightcurve_bis[:, 2]

            flux_source = min(flux_clean)
            good_points = np.where(flux_clean > flux_source)[0]

            while (np.std(time[good_points]) > 5) | (len(good_points) > 100):

                indexes = \
                    np.where((flux_clean[good_points] > np.median(flux_clean[good_points])) & (
                        errmag[good_points] <= max(0.1, 2.0 * np.mean(errmag[good_points]))))[0]

                if len(indexes) < 1:

                    break

                else:

                    good_points = good_points[indexes]

                    # gravity = (
                    #   np.median(time[good_points]), np.median(flux_clean[good_points]),
                    #    np.mean(errmag[good_points]))

                    # distances = np.sqrt((time[good_points] - gravity[0]) ** 2 / gravity[0] ** 2)

            to = np.median(time[good_points])
            max_flux = max(flux[good_points])
            to_estimations.append(to)
            maximum_flux_estimations.append(max_flux)
            errors_magnitude.append(np.mean(lightcurve_bis[good_points, 2]))

        except:

            time = lightcurve_magnitude[:, 0]
            flux = microltoolbox.magnitude_to_flux(lightcurve_magnitude[:, 1])
            to = np.median(time)
            max_flux = max(flux)
            to_estimations.append(to)
            maximum_flux_estimations.append(max_flux)

            errors_magnitude.append(mean_error_magnitude)

    to_guess = sum(np.array(to_estimations) / np.array(errors_magnitude) ** 2) / sum(
        1 / np.array(errors_magnitude) ** 2)
    survey = event.telescopes[0]
    lightcurve = survey.lightcurve_magnitude
    lightcurve = lightcurve[lightcurve[:, 0].argsort(), :]

    ## fs, uo, tE estimations only one the survey telescope


    time = lightcurve[:, 0]
    flux = microltoolbox.magnitude_to_flux(lightcurve[:, 1])
    errflux = microltoolbox.error_magnitude_to_error_flux(lightcurve[:, 2], flux)

    # fs estimation, no blend

    baseline_flux_0 = np.min(flux)
    baseline_flux = np.median(flux)

    while np.abs(baseline_flux_0 - baseline_flux) > 0.01 * baseline_flux:

        baseline_flux_0 = baseline_flux
        indexes = np.where((flux < baseline_flux))[0].tolist() + np.where(
            np.abs(flux - baseline_flux) < np.abs(errflux))[0].tolist()
        baseline_flux = np.median(flux[indexes])

        if len(indexes) < 100:
            baseline_flux = np.median(flux[flux.argsort()[:100]])
            break

    fs_guess = baseline_flux

    # uo estimation
    max_flux = maximum_flux_estimations[0]
    Amax = max_flux / fs_guess
    if (Amax < 1.0) | np.isnan(Amax):
        Amax = 1.1
    uo_guess = np.sqrt(-2 + 2 * np.sqrt(1 - 1 / (1 - Amax ** 2)))

    # tE estimations
    tE_guesses = []

    # Method 1 : flux(t_demi_amplification) = 0.5 * fs_guess * (Amax + 1)

    half_magnification = 0.5 * (Amax + 1)

    flux_demi_amplification = fs_guess * half_magnification

    index_plus = np.where((time > to_guess) & (flux < flux_demi_amplification))[0]
    index_moins = np.where((time < to_guess) & (flux < flux_demi_amplification))[0]

    if len(index_plus) != 0:

        if len(index_moins) != 0:
            t_demi_amplification = (time[index_plus[0]] - time[index_moins[-1]])
            tE_demi_amplification = t_demi_amplification / (
                2 * np.sqrt(-2 + 2 * np.sqrt(1 + 1 / (half_magnification ** 2 - 1)) - uo_guess ** 2))

            tE_guesses.append(tE_demi_amplification)

        else:
            t_demi_amplification = time[index_plus[0]] - to_guess
            tE_demi_amplification = t_demi_amplification / np.sqrt(
                -2 + 2 * np.sqrt(1 + 1 / (half_magnification ** 2 - 1)) - uo_guess ** 2)

            tE_guesses.append(tE_demi_amplification)
    else:

        if len(index_moins) != 0:
            t_demi_amplification = to_guess - time[index_moins[-1]]
            tE_demi_amplification = t_demi_amplification / np.sqrt(
                -2 + 2 * np.sqrt(1 + 1 / (half_magnification ** 2 - 1)) - uo_guess ** 2)

            tE_guesses.append(tE_demi_amplification)

    # Method 2 : flux(t_E) = fs_guess * (uo^+3)/[(uo^2+1)^0.5*(uo^2+5)^0.5]

    amplification_tE = (uo_guess ** 2 + 3) / ((uo_guess ** 2 + 1) ** 0.5 * np.sqrt(uo_guess ** 2 + 5))
    flux_tE = fs_guess * amplification_tE

    index_tE_plus = np.where((flux < flux_tE) & (time > to))[0]
    index_tE_moins = np.where((flux < flux_tE) & (time < to))[0]

    if len(index_tE_moins) != 0:
        index_tE_moins = index_tE_moins[-1]
        tE_moins = to_guess - time[index_tE_moins]

        tE_guesses.append(tE_moins)

    if len(index_tE_plus) != 0:
        index_tE_plus = index_tE_plus[0]
        tE_plus = time[index_tE_plus] - to_guess

        tE_guesses.append(tE_plus)

    # Method 3 : the first points before/after to_guess that reach the baseline. Very rough
    # approximation ot tE.

    index_tE_baseline_plus = np.where((time > to) & (np.abs(flux - fs_guess) < np.abs(errflux)))[0]
    index_tE_baseline_moins = np.where((time < to) & (np.abs(flux - fs_guess) < np.abs(errflux)))[0]

    if len(index_tE_baseline_plus) != 0:
        tEPlus = time[index_tE_baseline_plus[0]] - to_guess

        tE_guesses.append(tEPlus)

    if len(index_tE_baseline_moins) != 0:
        tEMoins = to_guess - time[index_tE_baseline_moins[-1]]

        tE_guesses.append(tEMoins)

    tE_guess = np.median(tE_guesses)

    # safety reason, unlikely
    if (tE_guess < 0.1) | np.isnan(tE_guess):
        tE_guess = 20.0

    # [to,uo,tE], fsource

    return [to_guess, uo_guess, tE_guess], fs_guess


def initial_guess_FSPL(event):
    """Function to find initial FSPL guess for Levenberg-Marquardt solver (method=='LM').
       This assumes no blending.

    :param object event: the event object on which you perform the fit on. More details on the
                            event module.

    :return: the FSPL guess for this event. A list of parameters associated to the FSPL model + the source flux of
                the survey telescope.
    :rtype: list,float
    """
    PSPL_guess, fs_guess = initial_guess_PSPL(event)
    # Dummy guess
    rho_guess = 0.05

    FSPL_guess = PSPL_guess + [rho_guess]

    # [to,uo,tE,rho], fsource
    return FSPL_guess, fs_guess


def initial_guess_DSPL(event):
    """Function to find initial DSPL guess for Levenberg-Marquardt solver (method=='LM').
       This assumes no blending.

       :param object event: the event object on which you perform the fit on. More details on the
                            event module.

       :return: the DSPL guess for this event. A list of parameters associated to the DSPL model + the source flux of
                the survey telescope.
       :rtype: list,float
    """
    PSPL_guess, fs_guess = initial_guess_PSPL(event)

    filters = [telescope.filter for telescope in event.telescopes]

    unique_filters = np.unique(filters)

    # Dummy guess
    delta_to_guess = 5  # days
    delta_uo_guess = 0.01
    q_flux_guess = 0.5

    DSPL_guess = PSPL_guess[:2] + [delta_to_guess] + [delta_uo_guess] + \
                 [PSPL_guess[2]] + [q_flux_guess] * len(unique_filters)

    # [to1,uo1,delta_to,uo2,tE,q_F_i], fsource
    return DSPL_guess, fs_guess


def differential_evolution_parameters_boundaries(model):
    """ This function define the parameters boundaries for a specific model.

       :param object model: a microlmodels object.

       :return: parameters_boundaries, a list of tuple containing parameters limits
       :rtype: list
    """

    minimum_observing_time_telescopes = [min(telescope.lightcurve_flux[:, 0]) - 0 for telescope in
                                         model.event.telescopes]
    maximum_observing_time_telescopes = [max(telescope.lightcurve_flux[:, 0]) + 0 for telescope in
                                         model.event.telescopes]

    to_boundaries = (min(minimum_observing_time_telescopes), max(maximum_observing_time_telescopes))
    delta_to_boundaries = (-150, 150)
    delta_uo_boundaries = (-1.0, 1.0)
    uo_boundaries = (-0.0, 1.0)
    tE_boundaries = (1.0, 500)
    rho_boundaries = (5 * 10 ** -5, 0.05)
    q_flux_boundaries = (0.001, 1.0)

    logs_boundaries = (-1.0, 1.0)
    logq_boundaries = (-6.0, 0.0)
    alpha_boundaries = (-np.pi, np.pi)

    piEN_boundaries = (-2.0, 2.0)
    piEE_boundaries = (-2.0, 2.0)
    XiEN_boundaries = (-2.0, 2.0)
    XiEE_boundaries = (-2.0, 2.0)

    dsdt_boundaries = (-1 * 10 ** -2, 1 * 10 ** -2)
    dalphadt_boundaries = (-1 * 10 ** -2, 1 * 10 ** -2)

    ra_xal_boundaries = [0,360]
    dec_xal_boundaries = [-90,90]
    period_xal_boundaries = [0.001,1000]
    ecc_xal_boundaries = [0,1]
    t_peri_xal_boundaries = to_boundaries


    # model_xallarap_boundaries = {'None': [], 'True': [(-2.0, 2.0), (-2.0, 2.0)]}

    # model_orbital_motion_boundaries = {'None': [], '2D': [], '3D': []}

    # model_source_spots_boundaries = {'None': []}

    Period = (0.0, 2.0)
    phis = (-np.pi, np.pi)
    amplitude = (0.0, 1.0)
    q_boundaries = (-2, 2)
    # Paczynski models boundaries
    if model.model_type == 'PSPL':
        parameters_boundaries = [to_boundaries, uo_boundaries, tE_boundaries]

    if model.model_type == 'FSPL':
        parameters_boundaries = [to_boundaries, uo_boundaries, tE_boundaries, rho_boundaries]

    if model.model_type == 'DSPL':
        parameters_boundaries = [to_boundaries, uo_boundaries, delta_to_boundaries,
                                 delta_uo_boundaries, tE_boundaries]
        filters = [telescope.filter for telescope in model.event.telescopes]

        unique_filters = np.unique(filters)

        parameters_boundaries += [q_flux_boundaries] * len(unique_filters)

    if model.model_type == 'PSBL':
        parameters_boundaries = [to_boundaries, uo_boundaries, tE_boundaries, logs_boundaries,
                                 logq_boundaries, alpha_boundaries]

    if (model.model_type == 'USBL') or (model.model_type == 'FSBL'):
        parameters_boundaries = [to_boundaries, uo_boundaries, tE_boundaries, rho_boundaries, logs_boundaries,
                                 logq_boundaries, alpha_boundaries]
        #fluxes = [(0,np.max(telescope.lightcurve_flux[:,1])) for telescope in model.event.telescopes]
        #blend = [(0,100) for telescope in model.event.telescopes]

        #for ind,telo in enumerate(model.event.telescopes):
             #parameters_boundaries+=[fluxes[ind], blend[ind]]

    if model.model_type == 'RRLyraePL':
        parameters_boundaries = [to_boundaries, uo_boundaries, tE_boundaries, Period]

        filters = [telescope.filter for telescope in model.event.telescopes]

        unique_filters = np.unique(filters)
        for i in range(model.number_of_harmonics):
            for j in unique_filters:
                parameters_boundaries += [amplitude]
                parameters_boundaries += [phis]

    if model.model_type == 'RRLyraeFS':
        parameters_boundaries = [to_boundaries, uo_boundaries, tE_boundaries, rho_boundaries, Period]

        filters = [telescope.filter for telescope in model.event.telescopes]

        unique_filters = np.unique(filters)
        for i in range(model.number_of_harmonics):
            for j in unique_filters:
                parameters_boundaries += [amplitude]
                parameters_boundaries += [phis]

    # Second order boundaries
    if model.parallax_model[0] != 'None':
        parameters_boundaries.append(piEN_boundaries)
        parameters_boundaries.append(piEE_boundaries)

    if model.xallarap_model != 'None':
        parameters_boundaries.append(XiEN_boundaries)
        parameters_boundaries.append(XiEE_boundaries)
        parameters_boundaries.append(ra_xal_boundaries)
        parameters_boundaries.append(dec_xal_boundaries)
        parameters_boundaries.append(period_xal_boundaries)

        if model.xallarap_model != 'circular':
            parameters_boundaries.append(ecc_xal_boundaries)
            parameters_boundaries.append(t_peri_xal_boundaries)

    if model.orbital_motion_model[0] != 'None':
        parameters_boundaries.append(dsdt_boundaries)
        parameters_boundaries.append(dalphadt_boundaries)
    # if source_spots

    return parameters_boundaries


def MCMC_parameters_initialization(parameter_key, parameters_dictionnary, parameters):
    """Generate a random parameter for the MCMC initialization.

        :param str parameter_key: the parameter on which we apply the function
        :param dict parameters_dictionnary: the dictionnary of parameters keys associared to the parameters input
        :param list parameters: a list of float which indicate the model parameters

        :return: a list containing the trial(s) associated to the parameter_key string
        :rtype: list of float
     """
    if ('to' in parameter_key) :
        epsilon = np.random.uniform(-0.001, 0.001)
        to_parameters_trial = parameters[parameters_dictionnary[parameter_key]] + epsilon

        return [to_parameters_trial]

    if 'fs' in parameter_key:
        epsilon = np.random.uniform(0.999, 1.001)

        fs_trial = parameters[parameters_dictionnary[parameter_key]] * epsilon
        g_trial = (1 + parameters[parameters_dictionnary[parameter_key] + 1]) / epsilon - 1

        return [fs_trial, g_trial]
        # return
    if ('g_' in parameter_key) or ('fb_' in parameter_key):
        return

    # if 'pi' in parameter_key:

    #    epsilon = np.random.uniform(0.9, 1.1)
    #    sign = np.random.choice([-1,1])

    #    pi_trial = sign*parameters[parameters_dictionnary[parameter_key]] * epsilon

    #    return [pi_trial]

    if 'rho' in parameter_key:
        epsilon = np.random.uniform(0.999, 1.001)
        rho_parameters_trial = parameters[parameters_dictionnary[parameter_key]] * epsilon
        return [rho_parameters_trial]

    if 'logs' in parameter_key:
        epsilon = np.random.uniform(-0.005, 0.005)

        logs_parameters_trial = parameters[parameters_dictionnary[parameter_key]] + epsilon
        return [logs_parameters_trial]
    if 'logq' in parameter_key:
        epsilon = np.random.uniform(-0.005, 0.005)

        logq_parameters_trial = parameters[parameters_dictionnary[parameter_key]] + epsilon
        return [logq_parameters_trial]
    epsilon = np.random.uniform(0.999, 1.001)
    all_other_parameter_trial = parameters[parameters_dictionnary[parameter_key]] * epsilon

    return [all_other_parameter_trial]



# This file was automatically generated by SWIG (http://www.swig.org).
# Version 3.0.10
#
# Do not make changes to this file unless you know what you are doing--modify
# the SWIG interface file instead.





from sys import version_info as _swig_python_version_info
if _swig_python_version_info >= (2, 7, 0):
    def swig_import_helper():
        import importlib
        pkg = __name__.rpartition('.')[0]
        mname = '.'.join((pkg, '_VBBinaryLensingLibrary')).lstrip('.')
        try:
            return importlib.import_module(mname)
        except ImportError:
            return importlib.import_module('_VBBinaryLensingLibrary')
    _VBBinaryLensingLibrary = swig_import_helper()
    del swig_import_helper
elif _swig_python_version_info >= (2, 6, 0):
    def swig_import_helper():
        from os.path import dirname
        import imp
        fp = None
        try:
            fp, pathname, description = imp.find_module('_VBBinaryLensingLibrary', [dirname(__file__)])
        except ImportError:
            import _VBBinaryLensingLibrary
            return _VBBinaryLensingLibrary
        if fp is not None:
            try:
                _mod = imp.load_module('_VBBinaryLensingLibrary', fp, pathname, description)
            finally:
                fp.close()
            return _mod
    _VBBinaryLensingLibrary = swig_import_helper()
    del swig_import_helper
else:
    import _VBBinaryLensingLibrary
del _swig_python_version_info
try:
    _swig_property = property
except NameError:
    pass  # Python < 2.2 doesn't have 'property'.

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_setattr_nondynamic(self, class_type, name, value, static=1):
    if (name == "thisown"):
        return self.this.own(value)
    if (name == "this"):
        if type(value).__name__ == 'SwigPyObject':
            self.__dict__[name] = value
            return
    method = class_type.__swig_setmethods__.get(name, None)
    if method:
        return method(self, value)
    if (not static):
        object.__setattr__(self, name, value)
    else:
        raise AttributeError("You cannot add attributes to %s" % self)


def _swig_setattr(self, class_type, name, value):
    return _swig_setattr_nondynamic(self, class_type, name, value, 0)


def _swig_getattr(self, class_type, name):
    if (name == "thisown"):
        return self.this.own()
    method = class_type.__swig_getmethods__.get(name, None)
    if method:
        return method(self)
    raise AttributeError("'%s' object has no attribute '%s'" % (class_type.__name__, name))


def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_method(set):
    def set_attr(self, name, value):
        if (name == "thisown"):
            return self.this.own(value)
        if hasattr(self, name) or (name == "this"):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add attributes to %s" % self)
    return set_attr


class doubleArray(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr

    def __init__(self, nelements):
        this = _VBBinaryLensingLibrary.new_doubleArray(nelements)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _VBBinaryLensingLibrary.delete_doubleArray
    __del__ = lambda self: None

    def __getitem__(self, index):
        return _VBBinaryLensingLibrary.doubleArray___getitem__(self, index)

    def __setitem__(self, index, value):
        return _VBBinaryLensingLibrary.doubleArray___setitem__(self, index, value)

    def cast(self):
        return _VBBinaryLensingLibrary.doubleArray_cast(self)
    frompointer = staticmethod(_VBBinaryLensingLibrary.doubleArray_frompointer)
doubleArray_swigregister = _VBBinaryLensingLibrary.doubleArray_swigregister
doubleArray_swigregister(doubleArray)

def doubleArray_frompointer(t):
    return _VBBinaryLensingLibrary.doubleArray_frompointer(t)
doubleArray_frompointer = _VBBinaryLensingLibrary.doubleArray_frompointer

_NP = _VBBinaryLensingLibrary._NP
class VBBinaryLensing(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr
    Tol = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_Tol_get, _VBBinaryLensingLibrary.VBBinaryLensing_Tol_set)
    satellite = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_satellite_get, _VBBinaryLensingLibrary.VBBinaryLensing_satellite_set)
    parallaxsystem = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_parallaxsystem_get, _VBBinaryLensingLibrary.VBBinaryLensing_parallaxsystem_set)
    nsat = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_nsat_get, _VBBinaryLensingLibrary.VBBinaryLensing_nsat_set)
    minannuli = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_minannuli_get, _VBBinaryLensingLibrary.VBBinaryLensing_minannuli_set)
    nannuli = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_nannuli_get, _VBBinaryLensingLibrary.VBBinaryLensing_nannuli_set)
    y_1 = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_y_1_get, _VBBinaryLensingLibrary.VBBinaryLensing_y_1_set)
    y_2 = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_y_2_get, _VBBinaryLensingLibrary.VBBinaryLensing_y_2_set)
    av = _swig_property(_VBBinaryLensingLibrary.VBBinaryLensing_av_get, _VBBinaryLensingLibrary.VBBinaryLensing_av_set)

    def PlotCrit(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_PlotCrit(self, arg2, arg3)

    def PrintCau(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_PrintCau(self, arg2, arg3)

    def BinaryMag0(self, *args):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinaryMag0(self, *args)

    def BinaryMag(self, *args):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinaryMag(self, *args)

    def BinaryMagDark(self, arg2, arg3, arg4, arg5, arg6, arg7, arg8):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinaryMagDark(self, arg2, arg3, arg4, arg5, arg6, arg7, arg8)

    def SetObjectCoordinates(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_SetObjectCoordinates(self, arg2, arg3)

    def PSPLCurve(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_PSPLCurve(self, arg2, arg3)

    def PSPLParallaxCurve(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_PSPLParallaxCurve(self, arg2, arg3)

    def ESPLCurve(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_ESPLCurve(self, arg2, arg3)

    def ESPLParallaxCurve(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_ESPLParallaxCurve(self, arg2, arg3)

    def BinaryLightCurve(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinaryLightCurve(self, arg2, arg3)

    def BinaryLightCurveW(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinaryLightCurveW(self, arg2, arg3)

    def BinaryLightCurveParallax(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinaryLightCurveParallax(self, arg2, arg3)

    def BinaryLightCurveOrbital(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinaryLightCurveOrbital(self, arg2, arg3)

    def BinSourceMag(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinSourceMag(self, arg2, arg3)

    def BinSourceParallaxMag(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinSourceParallaxMag(self, arg2, arg3)

    def BinSourceXallarapMag(self, arg2, arg3):
        return _VBBinaryLensingLibrary.VBBinaryLensing_BinSourceXallarapMag(self, arg2, arg3)

    def __init__(self):
        this = _VBBinaryLensingLibrary.new_VBBinaryLensing()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _VBBinaryLensingLibrary.delete_VBBinaryLensing
    __del__ = lambda self: None
VBBinaryLensing_swigregister = _VBBinaryLensingLibrary.VBBinaryLensing_swigregister
VBBinaryLensing_swigregister(VBBinaryLensing)

class annulus(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr
    bin = _swig_property(_VBBinaryLensingLibrary.annulus_bin_get, _VBBinaryLensingLibrary.annulus_bin_set)
    cum = _swig_property(_VBBinaryLensingLibrary.annulus_cum_get, _VBBinaryLensingLibrary.annulus_cum_set)
    Mag = _swig_property(_VBBinaryLensingLibrary.annulus_Mag_get, _VBBinaryLensingLibrary.annulus_Mag_set)
    err = _swig_property(_VBBinaryLensingLibrary.annulus_err_get, _VBBinaryLensingLibrary.annulus_err_set)
    f = _swig_property(_VBBinaryLensingLibrary.annulus_f_get, _VBBinaryLensingLibrary.annulus_f_set)
    nim = _swig_property(_VBBinaryLensingLibrary.annulus_nim_get, _VBBinaryLensingLibrary.annulus_nim_set)
    prev = _swig_property(_VBBinaryLensingLibrary.annulus_prev_get, _VBBinaryLensingLibrary.annulus_prev_set)
    next = _swig_property(_VBBinaryLensingLibrary.annulus_next_get, _VBBinaryLensingLibrary.annulus_next_set)

    def __init__(self):
        this = _VBBinaryLensingLibrary.new_annulus()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _VBBinaryLensingLibrary.delete_annulus
    __del__ = lambda self: None
annulus_swigregister = _VBBinaryLensingLibrary.annulus_swigregister
annulus_swigregister(annulus)

class _theta(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr
    th = _swig_property(_VBBinaryLensingLibrary._theta_th_get, _VBBinaryLensingLibrary._theta_th_set)
    maxerr = _swig_property(_VBBinaryLensingLibrary._theta_maxerr_get, _VBBinaryLensingLibrary._theta_maxerr_set)
    Mag = _swig_property(_VBBinaryLensingLibrary._theta_Mag_get, _VBBinaryLensingLibrary._theta_Mag_set)
    errworst = _swig_property(_VBBinaryLensingLibrary._theta_errworst_get, _VBBinaryLensingLibrary._theta_errworst_set)
    prev = _swig_property(_VBBinaryLensingLibrary._theta_prev_get, _VBBinaryLensingLibrary._theta_prev_set)
    next = _swig_property(_VBBinaryLensingLibrary._theta_next_get, _VBBinaryLensingLibrary._theta_next_set)

    def __init__(self, arg2):
        this = _VBBinaryLensingLibrary.new__theta(arg2)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _VBBinaryLensingLibrary.delete__theta
    __del__ = lambda self: None
_theta_swigregister = _VBBinaryLensingLibrary._theta_swigregister
_theta_swigregister(_theta)

class _thetas(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr
    first = _swig_property(_VBBinaryLensingLibrary._thetas_first_get, _VBBinaryLensingLibrary._thetas_first_set)
    last = _swig_property(_VBBinaryLensingLibrary._thetas_last_get, _VBBinaryLensingLibrary._thetas_last_set)
    length = _swig_property(_VBBinaryLensingLibrary._thetas_length_get, _VBBinaryLensingLibrary._thetas_length_set)

    def __init__(self):
        this = _VBBinaryLensingLibrary.new__thetas()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _VBBinaryLensingLibrary.delete__thetas
    __del__ = lambda self: None

    def insert(self, arg2):
        return _VBBinaryLensingLibrary._thetas_insert(self, arg2)
_thetas_swigregister = _VBBinaryLensingLibrary._thetas_swigregister
_thetas_swigregister(_thetas)

class complex(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr
    re = _swig_property(_VBBinaryLensingLibrary.complex_re_get, _VBBinaryLensingLibrary.complex_re_set)
    im = _swig_property(_VBBinaryLensingLibrary.complex_im_get, _VBBinaryLensingLibrary.complex_im_set)

    def __init__(self, *args):
        this = _VBBinaryLensingLibrary.new_complex(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _VBBinaryLensingLibrary.delete_complex
    __del__ = lambda self: None
complex_swigregister = _VBBinaryLensingLibrary.complex_swigregister
complex_swigregister(complex)

class _point(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr
    x1 = _swig_property(_VBBinaryLensingLibrary._point_x1_get, _VBBinaryLensingLibrary._point_x1_set)
    x2 = _swig_property(_VBBinaryLensingLibrary._point_x2_get, _VBBinaryLensingLibrary._point_x2_set)
    parab = _swig_property(_VBBinaryLensingLibrary._point_parab_get, _VBBinaryLensingLibrary._point_parab_set)
    ds = _swig_property(_VBBinaryLensingLibrary._point_ds_get, _VBBinaryLensingLibrary._point_ds_set)
    dJ = _swig_property(_VBBinaryLensingLibrary._point_dJ_get, _VBBinaryLensingLibrary._point_dJ_set)
    d = _swig_property(_VBBinaryLensingLibrary._point_d_get, _VBBinaryLensingLibrary._point_d_set)
    theta = _swig_property(_VBBinaryLensingLibrary._point_theta_get, _VBBinaryLensingLibrary._point_theta_set)

    def __init__(self, arg2, arg3, arg4):
        this = _VBBinaryLensingLibrary.new__point(arg2, arg3, arg4)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    next = _swig_property(_VBBinaryLensingLibrary._point_next_get, _VBBinaryLensingLibrary._point_next_set)
    prev = _swig_property(_VBBinaryLensingLibrary._point_prev_get, _VBBinaryLensingLibrary._point_prev_set)

    def __sub__(self, arg2):
        return _VBBinaryLensingLibrary._point___sub__(self, arg2)
    __swig_destroy__ = _VBBinaryLensingLibrary.delete__point
    __del__ = lambda self: None
_point_swigregister = _VBBinaryLensingLibrary._point_swigregister
_point_swigregister(_point)

class _curve(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr
    length = _swig_property(_VBBinaryLensingLibrary._curve_length_get, _VBBinaryLensingLibrary._curve_length_set)
    first = _swig_property(_VBBinaryLensingLibrary._curve_first_get, _VBBinaryLensingLibrary._curve_first_set)
    last = _swig_property(_VBBinaryLensingLibrary._curve_last_get, _VBBinaryLensingLibrary._curve_last_set)
    next = _swig_property(_VBBinaryLensingLibrary._curve_next_get, _VBBinaryLensingLibrary._curve_next_set)
    prev = _swig_property(_VBBinaryLensingLibrary._curve_prev_get, _VBBinaryLensingLibrary._curve_prev_set)
    partneratstart = _swig_property(_VBBinaryLensingLibrary._curve_partneratstart_get, _VBBinaryLensingLibrary._curve_partneratstart_set)
    partneratend = _swig_property(_VBBinaryLensingLibrary._curve_partneratend_get, _VBBinaryLensingLibrary._curve_partneratend_set)
    parabstart = _swig_property(_VBBinaryLensingLibrary._curve_parabstart_get, _VBBinaryLensingLibrary._curve_parabstart_set)

    def __init__(self, *args):
        this = _VBBinaryLensingLibrary.new__curve(*args)
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _VBBinaryLensingLibrary.delete__curve
    __del__ = lambda self: None

    def divide(self, arg2):
        return _VBBinaryLensingLibrary._curve_divide(self, arg2)

    def drop(self, arg2):
        return _VBBinaryLensingLibrary._curve_drop(self, arg2)

    def append(self, *args):
        return _VBBinaryLensingLibrary._curve_append(self, *args)

    def prepend(self, arg2, arg3):
        return _VBBinaryLensingLibrary._curve_prepend(self, arg2, arg3)

    def join(self, arg2):
        return _VBBinaryLensingLibrary._curve_join(self, arg2)

    def joinbefore(self, arg2):
        return _VBBinaryLensingLibrary._curve_joinbefore(self, arg2)

    def reverse(self):
        return _VBBinaryLensingLibrary._curve_reverse(self)

    def closest(self, arg2, arg3):
        return _VBBinaryLensingLibrary._curve_closest(self, arg2, arg3)

    def closest2(self, arg2, arg3):
        return _VBBinaryLensingLibrary._curve_closest2(self, arg2, arg3)

    def complement(self, arg2, arg3, arg4, arg5):
        return _VBBinaryLensingLibrary._curve_complement(self, arg2, arg3, arg4, arg5)
_curve_swigregister = _VBBinaryLensingLibrary._curve_swigregister
_curve_swigregister(_curve)

class _sols(object):
    thisown = _swig_property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc='The membership flag')
    __repr__ = _swig_repr
    length = _swig_property(_VBBinaryLensingLibrary._sols_length_get, _VBBinaryLensingLibrary._sols_length_set)
    first = _swig_property(_VBBinaryLensingLibrary._sols_first_get, _VBBinaryLensingLibrary._sols_first_set)
    last = _swig_property(_VBBinaryLensingLibrary._sols_last_get, _VBBinaryLensingLibrary._sols_last_set)

    def __init__(self):
        this = _VBBinaryLensingLibrary.new__sols()
        try:
            self.this.append(this)
        except __builtin__.Exception:
            self.this = this
    __swig_destroy__ = _VBBinaryLensingLibrary.delete__sols
    __del__ = lambda self: None

    def drop(self, arg2):
        return _VBBinaryLensingLibrary._sols_drop(self, arg2)

    def append(self, arg2):
        return _VBBinaryLensingLibrary._sols_append(self, arg2)

    def prepend(self, arg2):
        return _VBBinaryLensingLibrary._sols_prepend(self, arg2)

    def join(self, arg2):
        return _VBBinaryLensingLibrary._sols_join(self, arg2)
_sols_swigregister = _VBBinaryLensingLibrary._sols_swigregister
_sols_swigregister(_sols)

MR = _VBBinaryLensingLibrary.MR
MT = _VBBinaryLensingLibrary.MT
MAXIT = _VBBinaryLensingLibrary.MAXIT
MAXM = _VBBinaryLensingLibrary.MAXM

def rf(x, y, z):
    return _VBBinaryLensingLibrary.rf(x, y, z)
rf = _VBBinaryLensingLibrary.rf

def rd(x, y, z):
    return _VBBinaryLensingLibrary.rd(x, y, z)
rd = _VBBinaryLensingLibrary.rd

def rj(x, y, z, p):
    return _VBBinaryLensingLibrary.rj(x, y, z, p)
rj = _VBBinaryLensingLibrary.rj

def rc(x, y):
    return _VBBinaryLensingLibrary.rc(x, y)
rc = _VBBinaryLensingLibrary.rc

def ellf(phi, ak):
    return _VBBinaryLensingLibrary.ellf(phi, ak)
ellf = _VBBinaryLensingLibrary.ellf

def elle(phi, ak):
    return _VBBinaryLensingLibrary.elle(phi, ak)
elle = _VBBinaryLensingLibrary.elle

def ellpi(phi, en, ak):
    return _VBBinaryLensingLibrary.ellpi(phi, en, ak)
ellpi = _VBBinaryLensingLibrary.ellpi





# -*- coding: utf-8 -*-
"""
Created on Mon Nov  9 16:38:14 2015

@author: ebachelet
"""
from __future__ import division
from datetime import datetime
from collections import OrderedDict
import collections
import copy
import json

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator, MultipleLocator

import numpy as np
from astropy.time import Time
from scipy.stats.distributions import t as student

from pyLIMA import microltoolbox
from pyLIMA import microlstats
from pyLIMA import microlcaustics


plot_lightcurve_windows = 0.2
plot_residuals_windows = 0.2
MAX_PLOT_TICKS = 2
MARKER_SYMBOLS = np.array([['o', '.', '*', 'v', '^', '<', '>', 's', 'p', 'd', 'x']*10])
#plt.style.use('ggplot')


def pdf_output(fit, output_directory):
    from matplotlib.backends.backend_pdf import PdfPages
    with PdfPages(output_directory + fit.event.name + '.pdf') as pdf:
        figure_1 = fit.outputs.figure_lightcurve
        pdf.savefig(figure_1)

        figure_2 = fit.outputs.figure_geometry
        pdf.savefig(figure_2)

        if 'figure_distributions' in fit.outputs._fields:
            figure_3 = fit.outputs.figure_distributions
            pdf.savefig(figure_3)
        pdf_details = pdf.infodict()
        pdf_details['Title'] = fit.event.name + '_pyLIMA'
        pdf_details['Author'] = 'Produced by pyLIMA'
        pdf_details['Subject'] = 'A microlensing fit'

        pdf_details['CreationDate'] = datetime.today()

def statistical_outputs(fit) :
    """Compute statistics to estimate the fit quality

        :param object fit: a fit object. See the microlfits for more details.

        :return: a namedtuple containing the following attributes :

                  fit_parameters : an namedtuple object containing all the fitted parameters

                  fit_errors : an namedtuple object containing all the fitted parameters errors

                  fit_correlation_matrix : a numpy array representing the fitted parameters
                  correlation matrix

                  figure_lightcurve : a two matplotlib figure showing the data and model and the
                  correspoding residuals

        :rtype: object
    """
    fig_size = [15, 5]
    figure_stats = plt.figure(figsize=(fig_size[0], fig_size[1]))

    best_parameters = fit.fit_results
    best_model_pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(best_parameters)

    telescope_residuals = fit.all_telescope_residuals(best_model_pyLIMA_parameters)

    telescope_Kolmogorv_Smirnov_residuals_test = []
    telescope_Anderson_Darling_residuals_test = []
    telescope_Shapiro_Wilk_residuals_test = []

    telescope_chi2 = []
    telescope_chi2_sur_dof = []
    telescope_BIC = []
    telescope_AIC = []
    count = 0
    for telescope in fit.event.telescopes:

        residuals = telescope_residuals[count]

        telescope_residuals.append(residuals)

        Kolmogorov_Smirnov = microlstats.normal_Kolmogorov_Smirnov(residuals)

        Anderson_Darling = microlstats.normal_Anderson_Darling(residuals)


        Shapiro_Wilk = microlstats.normal_Shapiro_Wilk(residuals)

        telescope_Kolmogorv_Smirnov_residuals_test.append(Kolmogorov_Smirnov)
        telescope_Anderson_Darling_residuals_test.append(Anderson_Darling)
        telescope_Shapiro_Wilk_residuals_test.append(Shapiro_Wilk )

        chi2_sur_dof = microlstats.normalized_chi2((residuals**2).sum(), len(residuals),
                                                   len(fit.model.parameters_boundaries) + 2)
        BIC = 0.0
        AIC = 0.0

        telescope_chi2.append((residuals**2).sum())
        telescope_chi2_sur_dof.append(chi2_sur_dof)
        telescope_BIC.append(BIC)
        telescope_AIC.append(AIC)
        count += 1
    total_residuals = fit.residuals_LM(best_parameters)

    Kolmogorov_Smirnov = microlstats.normal_Kolmogorov_Smirnov(total_residuals)


    Anderson_Darling = microlstats.normal_Anderson_Darling(total_residuals)

    Shapiro_Wilk = microlstats.normal_Shapiro_Wilk(total_residuals)

    telescope_Kolmogorv_Smirnov_residuals_test.append(Kolmogorov_Smirnov)
    telescope_Anderson_Darling_residuals_test.append(Anderson_Darling)
    telescope_Shapiro_Wilk_residuals_test.append(Shapiro_Wilk)

    chi2_sur_dof = microlstats.normalized_chi2(best_parameters[-1], len(total_residuals),
                                               len(fit.model.parameters_boundaries)+2)
    BIC = microlstats.Bayesian_Information_Criterion(best_parameters[-1], len(total_residuals),
                                               len(fit.model.parameters_boundaries) + 2)
    AIC = microlstats.Akaike_Information_Criterion(best_parameters[-1], len(fit.model.parameters_boundaries) + 2)

    telescope_chi2.append(best_parameters[-1])
    telescope_chi2_sur_dof.append(chi2_sur_dof)
    telescope_BIC.append(BIC)
    telescope_AIC.append(AIC)




    raw_labels = [i.name for i in fit.event.telescopes]
    raw_labels += ['All site']
    column_labels = ['Kolmogorov-Smirnov\n(KS_stat,p_value)','Anderson-Darling\n(AD_stat,p value)','Shapiro-Wilk\n(SW_stat,p_value)','chi2','chi2_dof', 'BIC', 'AIC']
    table_val = []
    table_colors = []
    colors_dictionary = {0:'r',1:'y',2:'g'}

    for i in range(len(raw_labels)):
        table_val.append([np.round(telescope_Kolmogorv_Smirnov_residuals_test[i][:2],3),
                          np.round(telescope_Anderson_Darling_residuals_test[i][:2],3),
                          np.round(telescope_Shapiro_Wilk_residuals_test[i][:2],3),
                          np.round(telescope_chi2[i],3),np.round(telescope_chi2_sur_dof[i][0],3),
                          np.round(telescope_BIC[i],3),np.round(telescope_AIC[i],3)])

        table_colors.append([colors_dictionary[telescope_Kolmogorv_Smirnov_residuals_test[i][2]],
                             colors_dictionary[telescope_Anderson_Darling_residuals_test[i][2]],
                             colors_dictionary[telescope_Shapiro_Wilk_residuals_test[i][2]],
                             'w',
                             colors_dictionary[telescope_chi2_sur_dof[i][1]],
                             'w',
                             'w',
                             ])

    #table_val = np.round(table_val, 5).tolist()



    table_axes = figure_stats.add_subplot(111, frameon=False)

    the_table = table_axes.table(cellText=table_val,
                                 rowLabels=raw_labels, cellColours=table_colors,
                                 colLabels=column_labels, loc='center left')
    table_axes.get_yaxis().set_visible(False)
    table_axes.get_xaxis().set_visible(False)
    the_table.auto_set_font_size(False)
    the_table.set_fontsize(fig_size[0] * 3 / 4.0 / np.log10(len(fit.model.model_dictionnary.keys())))
    the_table.scale(0.75, 0.75)
    title = fit.model.event.name + ' : ' + fit.model.model_type
    figure_stats.suptitle(title, fontsize=30 * fig_size[0] / len(title))


def LM_outputs(fit):
    """Standard 'LM' and 'DE' outputs.

    :param object fit: a fit object. See the microlfits for more details.

    :return: a namedtuple containing the following attributes :

              fit_parameters : an namedtuple object containing all the fitted parameters

              fit_errors : an namedtuple object containing all the fitted parameters errors

              fit_correlation_matrix : a numpy array representing the fitted parameters
              correlation matrix

              figure_lightcurve : a two matplotlib figure showing the data and model and the
              correspoding residuals

    :rtype: object
    """
    # Change matplotlib default colors
    n = len(fit.event.telescopes)
    color = plt.cm.jet(np.linspace(0.01, 0.99, n))  # This returns RGBA; convert:
   # hexcolor = map(lambda rgb: '#%02x%02x%02x' % (rgb[0] * 255, rgb[1] * 255, rgb[2] * 255),
   #                tuple(color[:, 0:-1]))
    hexcolor = ['#'+format(int(i[0]*255),'x').zfill(2)+format(int(i[1]*255),'x').zfill(2)+
                format(int(i[2]*255),'x').zfill(2) for i in color]

    matplotlib.rcParams['axes.color_cycle'] = hexcolor
    #hexcolor[0] = '#000000'


    results = LM_parameters_result(fit)
    covariance_matrix = fit.fit_covariance
    errors = LM_fit_errors(fit)
    correlation_matrix = cov2corr(covariance_matrix)
    figure_lightcurve = LM_plot_lightcurves(fit)
    figure_trajectory = plot_LM_ML_geometry(fit)
    key_outputs = ['fit_parameters', 'fit_errors', 'fit_correlation_matrix', 'figure_lightcurve', 'figure_geometry']
    outputs = collections.namedtuple('Fit_outputs', key_outputs)

    values_outputs = [results, errors, correlation_matrix, figure_lightcurve, figure_trajectory]

    count = 0
    for key in key_outputs:
        setattr(outputs, key, values_outputs[count])
        count += 1

    return outputs


def MCMC_outputs(fit):
    """Standard 'LM' and 'DE' outputs.

    :param object fit: a fit object. See the microlfits for more details.

    :return: a namedtuple containing the following attributes :

             MCMC_chains : a numpy array containing all the parameters chains + the corresponding
             objective function.

             MCMC_correlations : a numpy array representing the fitted parameters correlation
             matrix from the MCMC chains

             figure_lightcurve : a two matplotlib subplot showing the data and 35 models and the
             residuals corresponding to the best model.

             figure_distributions : a multiple matplotlib subplot representing the parameters
             distributions (2D slice + histogram)

    :rtype: object
    """

    # Change matplotlib default colors

    n = len(fit.event.telescopes)
    color = plt.cm.jet(np.linspace(0.01, 0.99, n))  # This returns RGBA; convert:
    #hexcolor = map(lambda rgb: '#%02x%02x%02x' % (rgb[0] * 255, rgb[1] * 255, rgb[2] * 255),
    #              tuple(color[:, 0:-1]))
    hexcolor = ['#' + format(int(i[0] * 255), 'x').zfill(2) + format(int(i[1] * 255), 'x').zfill(2) +
                format(int(i[2] * 255), 'x').zfill(2) for i in color]
    matplotlib.rcParams['axes.color_cycle'] = hexcolor
    hexcolor[0] = '#000000'
    raw_chains = fit.MCMC_chains

    mcmc_chains = raw_chains


    best_probability = np.argmax(mcmc_chains[:, -1])

    # cut to 6 sigma for plots
    index = np.where(mcmc_chains[:, -1] > mcmc_chains[best_probability, -1] - 36)[0]
    best_chains = mcmc_chains[index]
    best_chains = best_chains[best_chains[:, -1].argsort(),]

    best_parameters = best_chains

    #best_parameters = best_parameters[index]
    best_parameters = best_parameters[best_parameters[:, -1].argsort(),]
    covariance_matrix = MCMC_covariance(mcmc_chains)
    correlation_matrix = cov2corr(covariance_matrix)

    figure_lightcurve = MCMC_plot_lightcurves(fit, best_chains)
    figure_distributions = MCMC_plot_parameters_distribution(fit, best_parameters)

    figure_geometry = plot_MCMC_ML_geometry(fit, best_chains)
    key_outputs = ['MCMC_chains', 'MCMC_correlations', 'figure_lightcurve', 'figure_distributions']
    outputs = collections.namedtuple('Fit_outputs', key_outputs)

    values_outputs = [mcmc_chains, correlation_matrix, figure_lightcurve, figure_distributions]

    count = 0
    for key in key_outputs:
        setattr(outputs, key, values_outputs[count])
        count += 1

    return outputs


def MCMC_compute_fs_g(fit, mcmc_chains):
    """ Compute the corresponding source flux fs and blending factor g corresponding to each mcmc
    chain.

    :param fit: a fit object. See the microlfits for more details.
    :param mcmc_chains: a numpy array representing the mcmc chains.
    :return: a numpy array containing the corresponding fluxes parameters
    :rtype: array_type

    """
    fluxes_chains = np.zeros((len(mcmc_chains), 2 * len(fit.event.telescopes)))
    for i in range(len(mcmc_chains)):

            fluxes = fit.find_fluxes(mcmc_chains[i], fit.model)
            fluxes_chains[i] = fluxes


    return fluxes_chains


def MCMC_plot_parameters_distribution(fit, mcmc_best):
    """ Plot the fit parameters distributions.
    Only plot the best mcmc_chains are plotted.
    :param fit: a fit object. See the microlfits for more details.
    :param mcmc_best: a numpy array representing the best (<= 6 sigma) mcmc chains.
    :return: a multiple matplotlib subplot representing the parameters distributions (2D slice +
    histogram)
    :rtype: matplotlib_figure
    """

    mcmc_string_format = np.array([str(i) for i in mcmc_best.tolist()])
    mcmc_to_plot = np.unique(mcmc_string_format)
    mcmc_unique = np.array([json.loads(i) for i in mcmc_to_plot])

    mcmc_unique = mcmc_unique[mcmc_unique[:, -1].argsort(),]
    dimensions =len(fit.model.model_dictionnary.keys())
    fig_size = [10, 10]

    max_plot_ticks = MAX_PLOT_TICKS

    figure_distributions, axes2 = plt.subplots(dimensions, dimensions, sharex='col', figsize=(fig_size[0], fig_size[1]))
    plt.subplots_adjust(top=0.9, bottom=0.15, left=0.10, right=0.9, wspace=0.3, hspace=0.3)
    count_i = 0

    for key_i in fit.model.model_dictionnary.keys()[: dimensions]:

        axes2[count_i, 0].set_ylabel(key_i, fontsize=5 * fig_size[0] * 3 / 2.0 / dimensions)
        axes2[-1, count_i].set_xlabel(key_i, fontsize=5 * fig_size[0] * 3 / 2.0 / dimensions)

        count_j = 0
        for key_j in fit.model.model_dictionnary.keys()[: dimensions]:

            axes2[count_i, count_j].ticklabel_format(useOffset=False, style='plain')
            axes2[count_i, count_j].ticklabel_format(useOffset=False, style='plain')

            if count_i == count_j:

                axes2[count_i, count_j].hist(mcmc_best[:, fit.model.model_dictionnary[key_i]], 100)
                axes2[count_i, count_j].xaxis.set_major_locator(MaxNLocator(max_plot_ticks, prune='both'))
                axes2[count_i, count_j].yaxis.set_major_locator(MaxNLocator(max_plot_ticks))
                axes2[count_i, count_j].tick_params(labelsize=5 * fig_size[0] * 3 / 3.0 / dimensions * 3 / 4.0)

            else:

                if count_j < count_i:

                    axes2[count_i, count_j].scatter(
                        mcmc_unique[:, fit.model.model_dictionnary[key_j]],
                        mcmc_unique[:, fit.model.model_dictionnary[key_i]],
                        c=mcmc_unique[:, -1],
                        edgecolor='None')

                    axes2[count_i, count_j].set_xlim(
                        [min(mcmc_unique[:, fit.model.model_dictionnary[key_j]]),
                         max(mcmc_unique[:, fit.model.model_dictionnary[key_j]])])
                    axes2[count_i, count_j].xaxis.set_major_locator(MaxNLocator(nbins=max_plot_ticks, prune='both'))

                    axes2[count_i, count_j].set_ylim(
                        [min(mcmc_unique[:, fit.model.model_dictionnary[key_i]]),
                         max(mcmc_unique[:, fit.model.model_dictionnary[key_i]])])

                    axes2[count_i, count_j].tick_params(labelsize=5 * fig_size[0] * 3 / 3.0 / dimensions * 3 / 4.0)

                else:

                    axes2[count_i, count_j].axis('off')

                if count_j == 0:

                    axes2[count_i, count_j].yaxis.set_major_locator(MaxNLocator(nbins=max_plot_ticks))
                else:

                    plt.setp(axes2[count_i, count_j].get_yticklabels(), visible=False)

            count_j += 1

        count_i += 1

    return figure_distributions


def MCMC_plot_lightcurves(fit, mcmc_best):
    """Plot 35 models from the mcmc_best sample. This is made to have  35 models equally spaced
    in term of objective funtion (~chichi)

    :param fit: a fit object. See the microlfits for more details.
    :param mcmc_best: a numpy array representing the best (<= 6 sigma) mcmc chains.
    :return: a two matplotlib subplot showing the data and 35 models and the residuals
    corresponding to the best model.
    :rtype: matplotlib_figure
    """

    figure_lightcurves, figure_axes = initialize_plot_lightcurve(fit)
    pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(mcmc_best[-1])
    MCMC_plot_align_data(fit, mcmc_best[0], figure_axes[0])

    model_panel_chichi = np.linspace(max(mcmc_best[:, -1]), min(mcmc_best[:, -1]), 35).astype(int)
    color_normalization = matplotlib.colors.Normalize(vmin=np.min(mcmc_best[:, -1]),
                                                      vmax=np.max(mcmc_best[:, -1]))
    color_map = matplotlib.cm.jet

    scalar_couleur_map = matplotlib.cm.ScalarMappable(cmap=color_map, norm=color_normalization)
    scalar_couleur_map.set_array([])

    min_time = min([min(i.lightcurve_magnitude[:, 0]) for i in fit.event.telescopes])
    max_time = max([max(i.lightcurve_magnitude[:, 0]) for i in fit.event.telescopes])

    time_of_model = np.linspace(min_time, max_time + 100, 30000)
    extra_time = np.linspace(pyLIMA_parameters.to - 2 * np.abs(pyLIMA_parameters.tE),
                                 pyLIMA_parameters.to + 2 * np.abs(pyLIMA_parameters.tE), 30000)

    time_of_model = np.sort(np.append(time_of_model, extra_time))
    reference_telescope = copy.copy(fit.event.telescopes[0])
    reference_telescope.lightcurve_magnitude = np.array(
        [time_of_model, [0] * len(time_of_model), [0] * len(time_of_model)]).T
    reference_telescope.lightcurve_flux = reference_telescope.lightcurve_in_flux()

    if fit.model.parallax_model[0] != 'None':
        reference_telescope.compute_parallax(fit.event, fit.model.parallax_model)

    for model_chichi in model_panel_chichi[0:]:
        indice = np.searchsorted(mcmc_best[:, -1], model_chichi) - 1

        MCMC_plot_model(fit, reference_telescope, mcmc_best[indice], mcmc_best[indice, -1], figure_axes[0],
                        scalar_couleur_map)

    colorbar = plt.colorbar(scalar_couleur_map, ax=figure_axes[0], orientation="horizontal")
    colorbar.locator = MaxNLocator(5)
    colorbar.formatter.set_useOffset(False)
    colorbar.update_ticks()

    figure_axes[0].text(0.01, 0.96, 'provided by pyLIMA', style='italic', fontsize=10,
                        transform=figure_axes[0].transAxes)

    figure_axes[0].invert_yaxis()
    MCMC_plot_residuals(fit, mcmc_best[0], figure_axes[1])

    return figure_lightcurves


def MCMC_plot_model(fit, reference_telescope, parameters, couleurs, figure_axes, scalar_couleur_map):
    """ Plot a  model to a given figure, with the color corresponding to the objective function
    of the model.

    :param fit: a fit object. See the microlfits for more details.
    :param parameters: the parameters [list] of the model you want to plot.
    :param couleurs: the values of the objective function for the model that match the color
    table scalar_couleur_map
    :param figure_axes: the axes where the plot are draw
    :param scalar_couleur_map: a matplotlib table that return a color given a scalar value (the
    objective function here)
    """

    pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(parameters)
    flux_model = fit.model.compute_the_microlensing_model(reference_telescope, pyLIMA_parameters)[0]

    magnitude_model = microltoolbox.flux_to_magnitude(flux_model)

    figure_axes.plot(reference_telescope.lightcurve_magnitude[:, 0], magnitude_model,
                     color=scalar_couleur_map.to_rgba(couleurs),
                     alpha=0.5)


def MCMC_plot_align_data(fit, parameters, plot_axe):
    """ Plot the data on the figure. Telescopes are aligned to the survey telescope (i.e number 0).

    :param fit: a fit object. See the microlfits for more details.
    :param parameters: the parameters [list] of the model you want to plot.
    :param plot_axe: the matplotlib axes where you plot the data
    """
    reference_telescope = fit.event.telescopes[0]
    pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(parameters)

    for index, telescope in enumerate(fit.event.telescopes):

        if telescope.name == reference_telescope:

            lightcurve = telescope.lightcurve_magnitude

        else:

            telescope_ghost = copy.copy(telescope)
            telescope_ghost.name = reference_telescope.name
            telescope_ghost.filter = reference_telescope.filter

            model_ghost = fit.model.compute_the_microlensing_model(telescope_ghost, pyLIMA_parameters)[0]
            model_telescope = fit.model.compute_the_microlensing_model(telescope, pyLIMA_parameters)[0]
            lightcurve_magnitude = align_telescope_lightcurve(telescope.lightcurve_flux, model_ghost, model_telescope)

        plot_axe.errorbar(lightcurve_magnitude[:, 0], lightcurve_magnitude[:, 1], yerr=lightcurve_magnitude[:, 2],
                          ls='None', marker=str(MARKER_SYMBOLS[0][index]), label=telescope.name)

    plot_axe.plot(fit.event.telescopes[0].lightcurve_magnitude[0, 0],
                  fit.event.telescopes[0].lightcurve_magnitude[0, 1],
                  '--k', label=fit.model.model_type, lw=1)
    plot_axe.legend(numpoints=1, bbox_to_anchor=(0.01, 0.90), loc=2, borderaxespad=0.)


def MCMC_plot_residuals(fit, parameters, ax):
    """Plot the data residual on the appropriate figure.

    :param fit: a fit object. See the microlfits for more details.
    :param parameters: the parameters [list] of the model you want to plot.
    :param ax: the matplotlib axes where you plot the data
    """

    for index, telescope in enumerate(fit.event.telescopes):
        time = telescope.lightcurve_flux[:, 0]
        flux = telescope.lightcurve_flux[:, 1]
        error_flux = telescope.lightcurve_flux[:, 2]
        err_mag = microltoolbox.error_flux_to_error_magnitude(error_flux, flux)

        pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(parameters)
        flux_model = fit.model.compute_the_microlensing_model(telescope, pyLIMA_parameters)[0]

        residuals = 2.5 * np.log10(flux_model / flux)
        ax.errorbar(time, residuals, yerr=err_mag, ls='None', marker=str(MARKER_SYMBOLS[0][index]))
    ax.set_ylim([-plot_residuals_windows, plot_residuals_windows])
    ax.invert_yaxis()
    ax.xaxis.get_major_ticks()[0].draw = lambda *args: None
    ax.ticklabel_format(useOffset=False, style='plain')


def LM_parameters_result(fit):
    """ Produce a namedtuple object containing the fitted parameters in the fit.fit_results.

    :param fit: a fit object. See the microlfits for more details.
    :param fit_parameters: a namedtuple object containing the fitted parameters.
    :rtype: object
    """

    fit_parameters = collections.namedtuple('Parameters', fit.model.model_dictionnary.keys())

    for parameter in fit.model.model_dictionnary.keys():
        setattr(fit_parameters, parameter, fit.fit_results[fit.model.model_dictionnary[parameter]])

    setattr(fit_parameters, 'chichi', fit.fit_results[-1])
    return fit_parameters


def MCMC_covariance(mcmc_chains):
    """ Estimate the covariance matrix from the mcmc_chains

    :param mcmc_chains: a numpy array representing the mcmc chains.
    :return : a numpy array representing the covariance matrix of your MCMC sampling.
    :rtype: array_like
    """
    esperances = []
    for i in range(mcmc_chains.shape[1] - 1):
        esperances.append(mcmc_chains[:, i] - np.median(mcmc_chains[:, i]))

    covariance_matrix = np.zeros((mcmc_chains.shape[1] - 1, mcmc_chains.shape[1] - 1))

    for i in range(mcmc_chains.shape[1] - 1):
        for j in np.arange(i, mcmc_chains.shape[1] - 1):
            covariance_matrix[i, j] = 1 / (len(mcmc_chains) - 1) * np.sum(
                esperances[i] * esperances[j])
            covariance_matrix[j, i] = 1 / (len(mcmc_chains) - 1) * np.sum(
                esperances[i] * esperances[j])

    return covariance_matrix


def LM_fit_errors(fit):
    """ Estimate the parameters errors from the fit.fit_covariance matrix.

    :param fit: a fit object. See the microlfits for more details.
    :return: a namedtuple object containing the square roots of parameters variance.
    :rtype: object
    """
    keys = ['err_' + parameter for parameter in fit.model.model_dictionnary.keys()]
    parameters_errors = collections.namedtuple('Errors_Parameters', keys)
    errors = fit.fit_covariance.diagonal() ** 0.5

    for i in fit.model.model_dictionnary.keys():
        setattr(parameters_errors, 'err_' + i, errors[fit.model.model_dictionnary[i]])

    return parameters_errors


def cov2corr(covariance_matrix):
    """Covariance matrix to correlation matrix.

    :param array_like covariance_matrix: a (square) numpy array representing the covariance matrix

    :return: a (square) numpy array representing the correlation matrix
    :rtype: array_like

    """

    covariance_diagonal = np.sqrt(covariance_matrix.diagonal())
    correlation_matrix = ((covariance_matrix.T / covariance_diagonal).T) / covariance_diagonal

    return correlation_matrix


def LM_plot_lightcurves(fit):
    """Plot the aligned datasets and the best fit on the first subplot figure_axes[0] and residuals
    on the second subplot figure_axes[1].

    :param object fit: a fit object. See the microlfits for more details.
    :return: a figure representing data+model and residuals.
    :rtype: matplotlib_figure

    """

    figure, figure_axes = initialize_plot_lightcurve(fit)
    LM_plot_model(fit, figure_axes[0])
    LM_plot_align_data(fit, figure_axes[0])
    LM_plot_residuals(fit, figure_axes[1])

    return figure


def LM_plot_parameters(fit):
    """ NOT USED
    """
    figure, axes = initialize_plot_parameters()

    return figure


def initialize_plot_lightcurve(fit):
    """Initialize the lightcurve plot.

    :param object fit: a fit object. See the microlfits for more details.

    :return: a matplotlib figure  and the corresponding matplotlib axes
    :rtype: matplotlib_figure,matplotlib_axes

    """
    fig_size = [10, 10]
    figure, figure_axes = plt.subplots(2, 1, sharex=True, gridspec_kw={'height_ratios': [3, 1]},
                                       figsize=(fig_size[0], fig_size[1]), dpi=75)
    plt.subplots_adjust(top=0.9, bottom=0.15, left=0.15, right=0.99, wspace=0.2, hspace=0.1)
    figure_axes[0].grid()
    figure_axes[1].grid()
    # fig_size = plt.rcParams["figure.figsize"]
    figure.suptitle(fit.event.name, fontsize=30 * fig_size[0] / len(fit.event.name))

    figure_axes[0].set_ylabel('Mag', fontsize=5 * fig_size[1] * 3 / 4.0)
    figure_axes[0].yaxis.set_major_locator(MaxNLocator(4))
    figure_axes[0].tick_params(axis='y', labelsize=3.5 * fig_size[1] * 3 / 4.0)

    figure_axes[1].set_xlabel('HJD', fontsize=5 * fig_size[0] * 3 / 4.0)
    figure_axes[1].xaxis.set_major_locator(MaxNLocator(6))
    figure_axes[1].yaxis.set_major_locator(MaxNLocator(4))
    figure_axes[1].xaxis.get_major_ticks()[0].draw = lambda *args: None

    figure_axes[1].ticklabel_format(useOffset=False, style='plain')
    figure_axes[1].set_ylabel('Residuals', fontsize=5 * fig_size[1] * 3 / 4.0)
    figure_axes[1].tick_params(axis='x', labelsize=3.5 * fig_size[0] * 3 / 4.0)
    figure_axes[1].tick_params(axis='y', labelsize=3.5 * fig_size[1] * 3 / 4.0)

    return figure, figure_axes


def initialize_plot_parameters(fit):
    """Initialize the parameters plot.

    :param object fit: a fit object. See the microlfits for more details.
    :return: a matplotlib figure  and the corresponding matplotlib axes.
    :rtype: matplotlib_figure,matplotlib_axes
    """

    fig_size = [10, 10]

    dimension_y = np.floor(len(fit.fits_result) / 3)
    dimension_x = len(fit.fits_result) - 3 * dimension_y

    figure, figure_axes = plt.subplots(dimension_x, dimension_y, figsize=(fig_size[0], fig_size[1]))
    plt.subplots_adjust(top=0.9, bottom=0.15, left=0.20, right=0.9, wspace=0.6, hspace=0.5)
    return figure, figure_axes


def LM_plot_model(fit, figure_axe):
    """Plot the microlensing model from the fit.

    :param object fit: a fit object. See the microlfits for more details.
    :param matplotlib_axes figure_axe: a matplotlib axes correpsonding to the figure.
    """
    pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(fit.fit_results)
    min_time = min([min(i.lightcurve_magnitude[:, 0]) for i in fit.event.telescopes])
    max_time = max([max(i.lightcurve_magnitude[:, 0]) for i in fit.event.telescopes])

    time = np.linspace(min_time, max_time + 100, 30000)
    extra_time = np.linspace(pyLIMA_parameters.to - 2 * np.abs(pyLIMA_parameters.tE),
                                 pyLIMA_parameters.to + 2 * np.abs(pyLIMA_parameters.tE), 30000)

    time = np.sort(np.append(time, extra_time))
    reference_telescope = copy.copy(fit.event.telescopes[0])
    reference_telescope.lightcurve_magnitude = np.array(
        [time, [0] * len(time), [0] * len(time)]).T
    reference_telescope.lightcurve_flux = reference_telescope.lightcurve_in_flux()

    if fit.model.parallax_model[0] != 'None':
        reference_telescope.compute_parallax(fit.event, fit.model.parallax_model)

    flux_model = fit.model.compute_the_microlensing_model(reference_telescope, pyLIMA_parameters)[0]
    magnitude = microltoolbox.flux_to_magnitude(flux_model)

    figure_axe.plot(time, magnitude, '--k', label=fit.model.model_type, lw=2)
    figure_axe.set_ylim(
        [min(magnitude) - plot_lightcurve_windows, max(magnitude) + plot_lightcurve_windows])
    figure_axe.set_xlim(
            [pyLIMA_parameters.to-3*np.abs(pyLIMA_parameters.tE), pyLIMA_parameters.to+3*np.abs(pyLIMA_parameters.tE)+100])

    figure_axe.invert_yaxis()
    figure_axe.text(0.01, 0.96, 'provided by pyLIMA', style='italic', fontsize=10,
                    transform=figure_axe.transAxes)


def LM_plot_residuals(fit, figure_axe):
    """Plot the residuals from the fit.

    :param object fit: a fit object. See the microlfits for more details.
    :param matplotlib_axes figure_axe: a matplotlib axes correpsonding to the figure.
    """

    for index, telescope in enumerate(fit.event.telescopes):
        time = telescope.lightcurve_flux[:, 0]
        flux = telescope.lightcurve_flux[:, 1]
        error_flux = telescope.lightcurve_flux[:, 2]
        err_mag = microltoolbox.error_flux_to_error_magnitude(error_flux, flux)

        pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(fit.fit_results)
        flux_model = fit.model.compute_the_microlensing_model(telescope, pyLIMA_parameters)[0]

        residuals = 2.5 * np.log10(flux_model / flux)

        figure_axe.errorbar(time, residuals, yerr=err_mag, ls='None', markersize=7.5,
                            marker=str(MARKER_SYMBOLS[0][index]),capsize=0.0)
    figure_axe.set_ylim([-plot_residuals_windows, plot_residuals_windows])
    figure_axe.invert_yaxis()
    figure_axe.yaxis.get_major_ticks()[-1].draw = lambda *args: None

    # xticks_labels = figure_axe.get_xticks()
    # figure_axe.set_xticklabels(xticks_labels, rotation=45)


def LM_plot_align_data(fit, figure_axe):
    """Plot the aligned data.

    :param object fit: a fit object. See the microlfits for more details.
    :param matplotlib_axes figure_axe: a matplotlib axes correpsonding to the figure.
    """

    normalised_lightcurves = microltoolbox.align_the_data_to_the_reference_telescope(fit)

    count = 0
    for index, telescope in enumerate(fit.event.telescopes):

        lightcurve = normalised_lightcurves[count]

        figure_axe.errorbar(lightcurve[:, 0], lightcurve[:, 1], yerr=lightcurve[:, 2], ls='None',
                            marker=str(MARKER_SYMBOLS[0][index]), markersize=7.5,capsize=0.0,
                            label=telescope.name)
        count += 1
    figure_axe.legend(numpoints=1, bbox_to_anchor=(0.01, 0.90), loc=2, borderaxespad=0.,
                      fancybox=True, framealpha=0.5)

def align_telescope_lightcurve(lightcurve_telescope_flux, model_ghost, model_telescope):
    """Align data to the survey telescope (i.e telescope 0).

    :param array_like lightcurve_telescope_mag: the survey telescope in magnitude
    :param float fs_reference: thce survey telescope reference source flux (i.e the fitted value)
    :param float g_reference: the survey telescope reference blending parameter (i.e the fitted
    value)
    :param float fs_telescope: the telescope source flux (i.e the fitted value)
    :param float g_reference: the telescope blending parameter (i.e the fitted value)

    :return: the aligned to survey lightcurve in magnitude
    :rtype: array_like
    """
    time = lightcurve_telescope_flux[:, 0]
    flux = lightcurve_telescope_flux[:, 1]
    error_flux = lightcurve_telescope_flux[:, 2]
    err_mag = microltoolbox.error_flux_to_error_magnitude(error_flux, flux)

    residuals = 2.5 * np.log10(model_telescope / flux)


    magnitude_normalised = microltoolbox.flux_to_magnitude(model_ghost)+residuals

    lightcurve_normalised = [time, magnitude_normalised, err_mag]

    lightcurve_mag_normalised = np.array(lightcurve_normalised).T

    return lightcurve_mag_normalised


def plot_LM_ML_geometry(fit):
    """Plot the lensing geometry (i.e source trajectory) and the table of best parameters.
    :param object fit: a fit object. See the microlfits for more details.
    :param list best_parameters: a list containing the model you want to plot the trajectory
    """

    # Limits of the plot

    figure_trajectory_xlimit = 1.5
    figure_trajectory_ylimit = 1.5

    best_parameters = fit.fit_results
    fig_size = [15, 5]
    figure_trajectory = plt.figure(figsize=(fig_size[0], fig_size[1]),dpi=75)

    figure_axes = figure_trajectory.add_subplot(121, aspect=1)
    plt.subplots_adjust(top=0.8, bottom=0.1, wspace=0.2, hspace=0.0)


    min_time = min([min(i.lightcurve_magnitude[:, 0]) for i in fit.event.telescopes])
    max_time = max([max(i.lightcurve_magnitude[:, 0]) for i in fit.event.telescopes])

    time = np.linspace(min_time, max_time + 100, 3000)

    reference_telescope = copy.copy(fit.event.telescopes[0])
    reference_telescope.lightcurve_magnitude = np.array(
        [time, [0] * len(time), [0] * len(time)]).T

    reference_telescope.lightcurve_flux = np.array(
        [time, [0] * len(time), [0] * len(time)]).T

    if fit.model.parallax_model[0] != 'None':
        reference_telescope.compute_parallax(fit.event, fit.model.parallax_model)

    pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(best_parameters)
    if 'BL' in fit.model.model_type:
        fit.model.find_origin(pyLIMA_parameters)
        to, uo = fit.model.uo_to_from_uc_tc(pyLIMA_parameters)


        trajectory_x,trajectory_y = fit.model.source_trajectory(reference_telescope, to, uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)
    else:

        trajectory_x, trajectory_y = fit.model.source_trajectory(reference_telescope, pyLIMA_parameters.to, pyLIMA_parameters.uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

    figure_axes.plot(trajectory_x, trajectory_y, 'b')

    # index_trajectory_limits = \
    #    np.where((np.abs(trajectory_x) < figure_trajectory_xlimit) & (np.abs(trajectory_y) < figure_trajectory_ylimit))[
    #        0]
    index_trajectory_limits = np.where((np.abs(time - pyLIMA_parameters.to) < 50))[0]


    if len(index_trajectory_limits) >= 3:
        midle = int(len(index_trajectory_limits) / 2)
        figure_axes.arrow(trajectory_x[index_trajectory_limits[midle]], trajectory_y[index_trajectory_limits[midle]],
                          trajectory_x[index_trajectory_limits[midle + 1]] - trajectory_x[
                              index_trajectory_limits[midle]],
                          trajectory_y[index_trajectory_limits[midle + 1]] - trajectory_y[
                              index_trajectory_limits[midle]],
                          head_width=0.01, head_length=0.01, color='b')

    if fit.model.model_type == 'DSPL':
        trajectory_x, trajectory_y = fit.model.source_trajectory(reference_telescope,
                                                                    pyLIMA_parameters.to + pyLIMA_parameters.delta_to,
                                                                    pyLIMA_parameters.uo + pyLIMA_parameters.delta_uo,
                                                                    pyLIMA_parameters.tE,
                                                                    pyLIMA_parameters)
        figure_axes.plot(trajectory_x, trajectory_y, 'r')
        index_trajectory_limits = np.where((np.abs(time - pyLIMA_parameters.to) < 50))[0]

        if len(index_trajectory_limits) >= 3:
            midle = int(len(index_trajectory_limits) / 2)
            figure_axes.arrow(trajectory_x[index_trajectory_limits[midle]],
                              trajectory_y[index_trajectory_limits[midle]],
                              trajectory_x[index_trajectory_limits[midle + 1]] - trajectory_x[
                                  index_trajectory_limits[midle]],
                              trajectory_y[index_trajectory_limits[midle + 1]] - trajectory_y[
                                  index_trajectory_limits[midle]],
                              head_width=0.1, head_length=0.2, color='r')

    if 'BL' not in fit.model.model_type:
        figure_axes.scatter(0, 0, s=10, c='k')
        einstein_ring = plt.Circle((0, 0), 1, fill=False, color='k', linestyle='--')
        figure_axes.add_artist(einstein_ring)

    if 'BL' in fit.model.model_type:
        regime, caustics, cc = microlcaustics.find_2_lenses_caustics_and_critical_curves(10**pyLIMA_parameters.logs,
                                                                                         10** pyLIMA_parameters.logq,
                                                                                         resolution=5000)
        for count, caustic in enumerate(caustics):

            try:
                figure_axes.plot(caustic.real, caustic.imag,lw=3,c='r')
                figure_axes.plot(cc[count].real, cc[count].imag, '--k')

            except AttributeError:
                pass


    if ('PS' not in fit.model.model_type) & ('DS' not in fit.model.model_type) & ('RRLyrae' not in fit.model.model_type):
        #index_source = np.where((trajectory_x ** 2 + trajectory_y ** 2) ** 0.5 < max(1, pyLIMA_parameters.uo + 0.1))[0][
        #   0]
        index_source  = np.argmin((trajectory_x ** 2 + trajectory_y ** 2) ** 0.5)
        source_disk = plt.Circle((trajectory_x[index_source], trajectory_y[index_source]), pyLIMA_parameters.rho,
                                 color='y')
        figure_axes.add_artist(source_disk)

    figure_axes.axis(
        [- figure_trajectory_xlimit, figure_trajectory_xlimit, - figure_trajectory_ylimit, figure_trajectory_ylimit])

    figure_axes.xaxis.set_major_locator(MaxNLocator(5))
    figure_axes.yaxis.set_major_locator(MaxNLocator(5))
    figure_axes.xaxis.get_major_ticks()[0].draw = lambda *args: None
    figure_axes.yaxis.get_major_ticks()[0].draw = lambda *args: None
    figure_axes.xaxis.get_major_ticks()[-1].draw = lambda *args: None
    figure_axes.yaxis.get_major_ticks()[-1].draw = lambda *args: None

    figure_axes.set_xlabel(r'$x(\theta_E)$', fontsize=50)
    figure_axes.set_ylabel(r'$y(\theta_E)$', fontsize=50)
    figure_axes.tick_params(axis='x', labelsize=25)
    figure_axes.tick_params(axis='y', labelsize=25)

    raw_labels = list(fit.model.model_dictionnary.keys()) + ['Chi^2']
    column_labels = ['Parameters', 'Errors']

    table_val = [fit.fit_results, (fit.fit_covariance.diagonal() ** 0.5).tolist() + [0.0]]
    table_val = np.round(table_val, 5).tolist()
    table_val = np.array(table_val).T.tolist()

    table_colors = []
    raw_colors = []
    last_color = 'dodgerblue'
    for i in range(len(table_val)):
        table_colors.append([last_color, last_color])
        raw_colors.append(last_color)

        if last_color == 'dodgerblue':

            last_color = 'white'

        else:

            last_color = 'dodgerblue'

    table_axes = figure_trajectory.add_subplot(122, frameon=False)

    the_table = table_axes.table(cellText=table_val, cellColours=table_colors, rowColours=raw_colors,
                                 rowLabels=raw_labels,
                                 colLabels=column_labels, loc='center left',
                                 rowLoc = 'left', colLoc='center',
                                 bbox=[0.0, -0.0, 1.0, 1.0]
                                 )
    table_axes.get_yaxis().set_visible(False)
    table_axes.get_xaxis().set_visible(False)
    #the_table.set_fontsize('large')
    #the_table.set_fontsize(fig_size[0] * 3 / 4.0 / np.log10(len(fit.model.model_dictionnary.keys())))
    #the_table.auto_set_font_size(False)
    #the_table.set_fontsize(fig_size[0] * 3 / 4.0 / np.log10(len(fit.model.model_dictionnary.keys())))
    #the_table.scale(0.9, 0.9)


    title = fit.model.event.name + ' : ' + fit.model.model_type
    figure_trajectory.suptitle(title, fontsize=30 * fig_size[0] / len(title))
    return figure_trajectory


def plot_MCMC_ML_geometry(fit, best_chains):
    """Plot the lensing geometry (i.e source trajectory) and the table of best parameters.
    :param object fit: a fit object. See the microlfits for more details.
    :param list best_parameters: a list containing the model you want to plot the trajectory
    """

    # Limits of the plot

    figure_trajectory_xlimit = 1.5
    figure_trajectory_ylimit = 1.5

    best_parameters = best_chains[0]
    fig_size = [15, 5]
    figure_trajectory = plt.figure(figsize=(fig_size[0], fig_size[1]))

    figure_axes = figure_trajectory.add_subplot(121, aspect=1, )
    plt.subplots_adjust(top=0.8, bottom=0.1, wspace=0.5, hspace=0.2)
    einstein_ring = plt.Circle((0, 0), 1, fill=False, color='k', linestyle='--')
    figure_axes.add_artist(einstein_ring)

    min_time = min([min(i.lightcurve_magnitude[:, 0]) for i in fit.event.telescopes])
    max_time = max([max(i.lightcurve_magnitude[:, 0]) for i in fit.event.telescopes])

    time = np.linspace(min_time, max_time + 100, 30000)

    reference_telescope = copy.copy(fit.event.telescopes[0])
    reference_telescope.lightcurve_magnitude = np.array(
        [time, [0] * len(time), [0] * len(time)]).T

    reference_telescope.lightcurve_flux = np.array(
        [time, [0] * len(time), [0] * len(time)]).T

    if fit.model.parallax_model[0] != 'None':
        reference_telescope.compute_parallax(fit.event, fit.model.parallax_model)

    pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(best_parameters)
    trajectory_x, trajectory_y = fit.model.source_trajectory(reference_telescope, pyLIMA_parameters.to,
                                                                pyLIMA_parameters.uo, pyLIMA_parameters.tE,
                                                                pyLIMA_parameters)

    figure_axes.plot(trajectory_x, trajectory_y, 'b')

    index_trajectory_limits = np.where((np.abs(time - pyLIMA_parameters.to) < 50))[0]
    if len(index_trajectory_limits) >= 3:
        midle = int(len(index_trajectory_limits) / 2)
        figure_axes.arrow(trajectory_x[index_trajectory_limits[midle]], trajectory_y[index_trajectory_limits[midle]],
                          trajectory_x[index_trajectory_limits[midle + 1]] - trajectory_x[
                              index_trajectory_limits[midle]],
                          trajectory_y[index_trajectory_limits[midle + 1]] - trajectory_y[
                              index_trajectory_limits[midle]],
                          head_width=0.1, head_length=0.2, color='b')

    if fit.model.model_type == 'DSPL':

        trajectory_x, trajectory_y = fit.model.source_trajectory(reference_telescope,
                                                                    pyLIMA_parameters.to + pyLIMA_parameters.delta_to,
                                                                    pyLIMA_parameters.uo + pyLIMA_parameters.delta_uo,
                                                                    pyLIMA_parameters.tE,
                                                                    pyLIMA_parameters)
        figure_axes.plot(trajectory_x, trajectory_y, 'r')
        index_trajectory_limits = np.where((np.abs(time - pyLIMA_parameters.to) < 50))[0]

        if len(index_trajectory_limits) >= 3:
            midle = int(len(index_trajectory_limits) / 2)
            figure_axes.arrow(trajectory_x[index_trajectory_limits[midle]],
                              trajectory_y[index_trajectory_limits[midle]],
                              trajectory_x[index_trajectory_limits[midle + 1]] - trajectory_x[
                                  index_trajectory_limits[midle]],
                              trajectory_y[index_trajectory_limits[midle + 1]] - trajectory_y[
                                  index_trajectory_limits[midle]],
                              head_width=0.1, head_length=0.2, color='r')

    if 'BL' not in fit.model.model_type:
        figure_axes.scatter(0, 0, s=10, c='k')
    if 'PS' not in fit.model.model_type:
        index_source = np.where((trajectory_x ** 2 + trajectory_y ** 2) ** 0.5 < max(1, pyLIMA_parameters.uo + 0.1))[0][
            0]
        source_disk = plt.Circle((trajectory_x[index_source], trajectory_y[index_source]), pyLIMA_parameters.rho,
                                 color='y')
        figure_axes.add_artist(source_disk)

    figure_axes.axis(
        [- figure_trajectory_xlimit, figure_trajectory_xlimit, - figure_trajectory_ylimit, figure_trajectory_ylimit])

    raw_labels = list(fit.model.model_dictionnary.keys())
    column_labels = ['Parameters 16%', 'Parameters 50%', 'Parameters 84%']
    table_val = []
    for i in range(len(fit.model.model_dictionnary.keys())):
        table_val.append([np.percentile(best_chains[:, i], 16), np.percentile(best_chains[:, i], 50),
                          np.percentile(best_chains[:, i], 84)])

    table_val = np.round(table_val, 5).tolist()

    table_colors = []
    raw_colors = []
    last_color = 'dodgerblue'
    for i in range(len(table_val)):

        table_colors.append([last_color, last_color, last_color])
        raw_colors.append(last_color)

        if last_color == 'dodgerblue':

            last_color = 'white'

        else:

            last_color = 'dodgerblue'

    table_axes = figure_trajectory.add_subplot(122, frameon=False)

    the_table = table_axes.table(cellText=table_val, cellColours=table_colors, rowColours=raw_colors,
                                 rowLabels=raw_labels,
                                 colLabels=column_labels, loc='center left')
    table_axes.get_yaxis().set_visible(False)
    table_axes.get_xaxis().set_visible(False)
    the_table.auto_set_font_size(False)
    the_table.set_fontsize(fig_size[0] * 3 / 4.0 / np.log10(len(fit.model.model_dictionnary.keys())))
    the_table.scale(0.75, 0.75)
    title = fit.model.event.name + ' : ' + fit.model.model_type
    figure_trajectory.suptitle(title, fontsize=30 * fig_size[0] / len(title))

    return figure_trajectory



# -*- coding: utf-8 -*-
"""
Created on Mon May 23 17:18:15 2016

@author: ebachelet
"""

import numpy as np
import copy

# magnitude reference
MAGNITUDE_CONSTANT = 27.4


def chichi(residuals_fn, fit_process_parameters):
    """Return the chi^2 .


    :param func residuals_fn: a function which compute the residuals
    :param list fit_process_parameters: the model parameters ingested by the correpsonding
                                        fitting routine.

    :returns: the chi^2

    :rtype: float
    """

    residuals = residuals_fn(fit_process_parameters)
    _chichi = np.sum(residuals ** 2)

    return _chichi


def magnitude_to_flux(magnitude):
    """ Transform the injected magnitude to the the corresponding flux.

    :param array_like magnitude: the magnitude you want to transform.

    :return: the transformed magnitude in flux unit
    :rtype: array_like
    """

    flux = 10 ** ((MAGNITUDE_CONSTANT - magnitude) / 2.5)

    return flux


def flux_to_magnitude(flux):
    """ Transform the injected flux to the the corresponding magnitude.

    :param array_like flux: the flux you want to transform.

    :return: the transformed magnitude
    :rtype: array_like
    """

    mag = MAGNITUDE_CONSTANT - 2.5 * np.log10(flux)

    return mag


def error_magnitude_to_error_flux(error_magnitude, flux):
    """ Transform the injected magnitude error to the the corresponding error in flux.

    :param array_like error_magnitude: the magnitude errors measurements you want to transform.
    :param array_like flux: the fluxes corresponding to these errors

    :return: the transformed errors in flux units
    :rtype: array_like
    """

    error_flux = np.abs(-error_magnitude * flux * np.log(10) / 2.5)

    return error_flux


def error_flux_to_error_magnitude(error_flux, flux):
    """ Transform the injected flux error to the the corresponding error in magnitude.

    :param array_like error_flux: the flux errors measurements you want to transform.
    :param array_like flux: the fluxes corresponding to these errors

    :return: the transformed errors in magnitude
    :rtype: array_like
    """
    error_magnitude = np.abs(-2.5 * error_flux / (flux * np.log(10)))

    return error_magnitude


def align_the_data_to_the_reference_telescope(fit):
    """Align data to the survey telescope (i.e telescope 0). Used to plot fit results. Ugly microlensing alignement....

    :param object fit: a microlfits object

    :return: the aligned to survey lightcurve in magnitude
    :rtype: array_like
    """

    reference_telescope = fit.event.telescopes[0]
    pyLIMA_parameters = fit.model.compute_pyLIMA_parameters(fit.fit_results)

    normalised_lightcurve = []
    for telescope in fit.event.telescopes:

        if telescope.name == reference_telescope.name:

            lightcurve = telescope.lightcurve_magnitude

        else:

            telescope_ghost = copy.copy(telescope)
            telescope_ghost.name = reference_telescope.name
            telescope_ghost.filter = reference_telescope.filter
            # import pdb;
            # pdb.set_trace()
            model_ghost = fit.model.compute_the_microlensing_model(telescope_ghost, pyLIMA_parameters)[0]
            model_telescope = fit.model.compute_the_microlensing_model(telescope, pyLIMA_parameters)[0]

            time = telescope.lightcurve_flux[:, 0]
            flux = telescope.lightcurve_flux[:, 1]
            error_flux = telescope.lightcurve_flux[:, 2]
            err_mag = error_flux_to_error_magnitude(error_flux, flux)

            residuals = 2.5 * np.log10(model_telescope / flux)

            magnitude_normalised = flux_to_magnitude(model_ghost) + residuals

            lightcurve_normalised = [time, magnitude_normalised, err_mag]

            lightcurve_mag_normalised = np.array(lightcurve_normalised).T

            lightcurve = lightcurve_mag_normalised

        normalised_lightcurve.append(lightcurve)

    return normalised_lightcurve



import numpy as np
import astropy
from astropy.coordinates import SkyCoord, EarthLocation, AltAz, get_sun, get_moon
from astropy.time import Time

import matplotlib.pyplot as plt

from pyLIMA import microlmodels
from pyLIMA import microltoolbox
from pyLIMA import telescopes
from pyLIMA import event

RED_NOISE = 'Yes'
SOURCE_MAGNITUDE = [14, 22]
BLEND_LIMITS = [0, 1]


def moon_illumination(sun, moon):
    """The moon illumination expressed as a percentage.

            :param astropy sun: the sun ephemeris
            :param astropy moon: the moon ephemeris

            :return: a numpy array indicated the moon illumination.

            :rtype: array_like

    """

    geocentric_elongation = sun.separation(moon).rad
    selenocentric_elongation = np.arctan2(sun.distance * np.sin(geocentric_elongation),
                                          moon.distance - sun.distance * np.cos(geocentric_elongation))

    illumination = (1 + np.cos(selenocentric_elongation)) / 2.0

    return illumination


def poisson_noise(flux):
    """The Poisson noise.

        :param array_like flux: the observed flux

        :return: a numpy array which represents the Poisson noise,

        :rtype: array_like

    """
    error_flux = flux ** 0.5

    return error_flux


def noisy_observations(flux, error_flux):
    """Add Poisson noise to observations.

        :param array_like flux: the observed flux
        :param array_like error_flux: the error on observed flux

        :return: a numpy array which represents the observed noisy flux

        :rtype: array_like

    """
    try:

        flux_observed = np.random.normal(flux, error_flux)

    except:

        flux_observed = flux

    return flux_observed


def time_simulation(time_start, time_end, sampling, bad_weather_percentage):
    """ Simulate observing time during the observing windows, rejecting windows with bad weather.

    :param float time_start: the start of observations in JD
    :param float time_end: the end of observations in JD
    :param float sampling: the number of points observed per hour.
    :param float bad_weather_percentage: the percentage of bad nights

    :return: a numpy array which represents the time of observations

    :rtype: array_like

    """

    total_number_of_days = int(time_end - time_start)
    time_step_observations = sampling / 24.0
    number_of_day_exposure = np.floor(
        1.0 / time_step_observations)  # less than expected total, more likely in a telescope :)
    night_begin = time_start

    time_observed = []
    for i in range(total_number_of_days):

        good_weather = np.random.uniform(0, 1)

        if good_weather > bad_weather_percentage:
            random_begin_of_the_night = 0
            night_end = night_begin + 1.0
            time_observed += np.linspace(night_begin + time_step_observations + random_begin_of_the_night, night_end,
                                         number_of_day_exposure).tolist()

        night_begin += 1

    time_of_observations = np.array(time_observed)

    return time_of_observations


def red_noise(time):
    """ Simulate red moise as a sum of 10 low amplitudes/period sinusoidals.

        :param array_like time: the time in JD where you simulate red noise

        :return: a numpy array which represents the red noise

        :rtype: array_like

    """

    red_noise_amplitude = np.random.random_sample(10) * 0.5 / 100
    red_noise_period = np.random.random_sample(10)
    red_noise_phase = np.random.random_sample(10) * 2 * np.pi

    red_noise = 0
    for j in range(10):
        red_noise += np.sin(2 * np.pi * time / red_noise_period[j] + red_noise_phase[j]) * red_noise_amplitude[j]

    return red_noise


def simulate_a_microlensing_event(name='Microlensing pyLIMA simulation', ra=270, dec=-30):
    """ Simulate a microlensing event. More details in the event module.

        :param str name:  the name of the event. Default is 'Microlensing pyLIMA simulation'
        :param float ra: the right ascension in degrees of your simulation. Default is 270.
        :param float dec: the declination in degrees of your simulation. Default is -30.


        :return: a event object

        :rtype: object
    """

    fake_event = event.Event()
    fake_event.name = name
    fake_event.ra = ra
    fake_event.dec = dec

    return fake_event


def simulate_a_telescope(name, event, time_start, time_end, sampling, location, filter, uniform_sampling=False,
                         altitude=0, longitude=0, latitude=0, spacecraft_name=None, bad_weather_percentage=0.0,
                         minimum_alt=20, moon_windows_avoidance=20, maximum_moon_illumination=100.0):
    """ Simulate a telescope. More details in the telescopes module. The observations simulation are made for the
        full time windows, then limitation are applied :
            - Sun has to be below horizon : Sun< -18
            - Moon has to be more than the moon_windows_avoidance distance from the target
            - Observations altitude of the target have to be bigger than minimum_alt

    :param str name:  the name of the telescope.
    :param object event: the microlensing event you look at
    :param float time_start: the start of observations in JD
    :param float time_end: the end of observations in JD
    :param float sampling: the hour sampling.
    :param str location: the location of the telescope.
    :param str filter: the filter used for observations
    :param boolean uniform_sampling: set it to True if you want no bad weather, no moon avoidance etc....

    :param float altitude: the altitude in meters if the telescope
    :param float longitude: the longitude in degree of the telescope location
    :param float latitude: the latitude in degree of the telescope location

    :param str spacecraft_name: the name of your satellite according to JPL horizons

    :param float bad_weather_percentage: the percentage of bad nights
    :param float minimum_alt: the minimum altitude ini degrees that your telescope can go to.
    :param float moon_windows_avoidance: the minimum distance in degrees accepted between the target and the Moon
    :param float maximum_moon_illumination: the maximum Moon brightness you allow in percentage
    :return: a telescope object
    :rtype: object
    """

    # fake lightcurve
    if (uniform_sampling == False) & (location != 'Space'):
        earth_location = EarthLocation(lon=longitude * astropy.units.deg,
                                       lat=latitude * astropy.units.deg,
                                       height=altitude * astropy.units.m)

        target = SkyCoord(event.ra, event.dec, unit='deg')

        minimum_sampling = min(4.0, sampling)
        ratio_sampling = np.round(sampling / minimum_sampling)

        time_of_observations = time_simulation(time_start, time_end, minimum_sampling,
                                               bad_weather_percentage)

        time_convertion = Time(time_of_observations, format='jd').isot

        telescope_altaz = target.transform_to(AltAz(obstime=time_convertion, location=earth_location))
        altazframe = AltAz(obstime=time_convertion, location=earth_location)
        Sun = get_sun(Time(time_of_observations, format='jd')).transform_to(altazframe)
        Moon = get_moon(Time(time_of_observations, format='jd')).transform_to(altazframe)
        Moon_illumination = moon_illumination(Sun, Moon)
        Moon_separation = target.separation(Moon)
        observing_windows = np.where((telescope_altaz.alt > minimum_alt * astropy.units.deg)
                                     & (Sun.alt < -18 * astropy.units.deg)
                                     & (Moon_separation > moon_windows_avoidance * astropy.units.deg)
                                     & (Moon_illumination < maximum_moon_illumination)
                                     )[0]

        time_of_observations = time_of_observations[observing_windows]


    else:

        time_of_observations = np.arange(time_start, time_end, sampling / (24.0))

    lightcurveflux = np.ones((len(time_of_observations), 3)) * 42
    lightcurveflux[:, 0] = time_of_observations

    telescope = telescopes.Telescope(name=name, camera_filter=filter, light_curve_flux=lightcurveflux,
                                     location=location, spacecraft_name=spacecraft_name)

    return telescope


def simulate_a_microlensing_model(event, model='PSPL', args=(), parallax=['None', 0.0], xallarap=['None', 0.0],
                                  orbital_motion=['None', 0.0], source_spots='None'):
    """ Simulate a a microlensing model.

    :param object event: the microlensing event you look at. More details in event module
    :param str model: the microlensing model you want. Default is 'PSPL'. More details in microlmodels module
    :param array_like parallax: the parallax effect you want to add. Default is no parallax.
                                     More details in microlmodels module
    :param array_like xallarap: the xallarap effect you want to add. Default is no parallax.
                                     More details in microlmodels module
    :param str source_spots: If you want to add source spots. Default is no source_spots.
                                     More details in microlmodels module

    :return: a microlmodel object
    :rtype: object
    """

    fake_model = microlmodels.create_model(model, event, args, parallax, xallarap,
                                           orbital_motion, source_spots)
    fake_model.define_model_parameters()

    return fake_model


def simulate_microlensing_model_parameters(model):
    """ Simulate parameters given the desired model. Parameters are selected in uniform distribution inside
        parameters_boundaries given by the microlguess modules. The exception is 'to' where it is selected
        to enter inside telescopes observations.

        :param object event: the microlensing event you look at. More details in event module


        :return: fake_parameters, a set of parameters
        :rtype: list
    """

    fake_parameters = []

    for key in list(model.pyLIMA_standards_dictionnary.keys())[:len(model.parameters_boundaries)]:

        if key == 'to':

            minimum_acceptable_time = max([min(i.lightcurve_flux[:, 0]) for i in model.event.telescopes])
            maximum_acceptable_time = min([max(i.lightcurve_flux[:, 0]) for i in model.event.telescopes])

            fake_parameters.append(np.random.uniform(minimum_acceptable_time, maximum_acceptable_time))

        else:

            boundaries = model.parameters_boundaries[model.pyLIMA_standards_dictionnary[key]]
            fake_parameters.append(np.random.uniform(boundaries[0], boundaries[1]))

    if model.model_type == 'FSPL':
        if np.abs(fake_parameters[1]) > 0.1:
            fake_parameters[1] /= 100
        if np.abs(fake_parameters[1] / fake_parameters[3]) > 10:
            fake_parameters[1] = np.abs(fake_parameters[1]) * np.random.uniform(0, fake_parameters[3])

    if model.model_type == 'DSPL':

        if np.abs(fake_parameters[2]) > 100:
            fake_parameters[2] = np.random.uniform(10, 15)

    return fake_parameters


def simulate_fluxes_parameters(list_of_telescopes):
    """ Simulate flux parameters (magnitude_source , g) for the telescopes. More details in microlmodels module

    :param list list_of_telescopes: a list of telescopes object

    :return: fake_fluxes parameters, a set of fluxes parameters
    :rtype: list

    """

    fake_fluxes_parameters = []

    for telescope in list_of_telescopes:
        magnitude_source = np.random.uniform(SOURCE_MAGNITUDE[0], SOURCE_MAGNITUDE[1])
        flux_source = microltoolbox.magnitude_to_flux(magnitude_source)
        blending_ratio = np.random.uniform(BLEND_LIMITS[0], BLEND_LIMITS[1])

        fake_fluxes_parameters.append(flux_source)
        fake_fluxes_parameters.append(blending_ratio)

    return fake_fluxes_parameters


def simulate_lightcurve_flux(model, pyLIMA_parameters, red_noise_apply='Yes'):
    """ Simulate the flux of telescopes given a model and a set of parameters.
    It updates straight the telescopes object inside the given model.

    :param object model: the microlensing model you desire. More detail in microlmodels.
    :param object pyLIMA_parameters: the parameters used to simulate the flux.
    :param str red_noise_apply: to include or not red_noise

    """

    count = 0

    for telescope in model.event.telescopes:

        theoritical_flux = model.compute_the_microlensing_model(telescope, pyLIMA_parameters)[0]
        if np.min(theoritical_flux > 0):
            pass
        else:
            import pdb;
            pdb.set_trace()
        flux_error = poisson_noise(theoritical_flux)
        observed_flux = noisy_observations(theoritical_flux, flux_error)

        if red_noise_apply == 'Yes':
            red = red_noise(telescope.lightcurve_flux[:, 0])

            redded_flux = (1 - np.log(10) / 2.5 * red) * observed_flux
            error_on_redded_flux = poisson_noise(redded_flux)

        else:

            redded_flux = observed_flux
            error_on_redded_flux = poisson_noise(redded_flux)

        telescope.lightcurve_flux[:, 1] = redded_flux
        telescope.lightcurve_flux[:, 2] = error_on_redded_flux

        telescope.lightcurve_magnitude = telescope.lightcurve_in_magnitude()

        count += 1



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 27 16:39:32 2015

@author: ebachelet
"""

from __future__ import division

import sys
import numpy as np

from pyLIMA import microlfits


class EventException(Exception):
    pass


class Event(object):
    """
    ######## Event module ########

    This module create an event object with the informations (attributes) needed for the fits.

    Attributes :

         kind : type of event. In general, should be 'Microlensing' (default)

         name : name of the event. Should be a string. Default is 'Sagittarius A*'

         ra : right ascension of the event (J2000). Should be a float in degree between 0.0 and
         360.0. Default is ra of Sagittarius A, :math:`\\alpha` = 266.416792 from Baker & Sramek
         1999ApJ...524..805B

         dec : declination of the event (J2000). Should be a float in degree between -90 and 90.
         Default is dec of Sagittarius A, :math:`\\delta` = -29.007806 from Baker & Sramek
         1999ApJ...524..805B

         Teff : effective temperature of the star in Kelvin. Should be a float. Default is 5000.0 K

         log_g : surface gravity in log10 cgs unit. Should be a float. Default is 4.5

         telescopes : list of telescopes names (strings). Default is an empty list. Have to be
         fill with some telescopes class instances.

         survey : the reference telescope. Has to be a string, default is 'None'.

         fits : list of microlfits objects.
         
    """

    def __init__(self):
        """ Initialization of the attributes described above. """

        self.kind = 'Microlensing'
        self.name = 'Sagittarius A*'
        self.ra = 266.416792
        self.dec = -29.007806
        self.Teff = 5000  # Kelvins
        self.log_g = 4.5
        self.telescopes = []
        self.survey = 'None'
        self.fits = []

    def fit(self, model, method, DE_population_size=10, flux_estimation_MCMC='MCMC', fix_parameters_dictionnary=None,
            grid_resolution=10, computational_pool=None, binary_regime=None):
        """Function to fit the event with a model and a method.


        :param model: the model you want to fit. More details in the microlfits module

        :param method: the fitting method you want to use. Has to be a string in the
        available_methods parameter:

            'LM' : Levenberg-Marquardt algorithm

            'DE' : Differential Evolution algorithm

            'MCMC' : Monte-Carlo Markov Chain algorithm

            More details in the microlfits module

        A microlfits object is added in the event.fits list. For example, if you request two fits,
        you will obtain :

        event.fits=[fit1,fit2]

        More details in the microlfits module.

        """
        available_kind = ['Microlensing']
        available_methods = ['LM', 'DE', 'MCMC', 'GRIDS', 'TRF']

        if self.kind not in available_kind:
            print('ERROR : No possible fit yet for a non microlensing event, sorry :(')
            raise EventException('Can not fit this event kind')

        if method not in available_methods:
            print('ERROR : Wrong method request, has to be a string selected between ' + \
                  ' or '.join(available_methods) + '')
            raise EventException('Wrong fit method request')

        fit = microlfits.MLFits(self)
        fit.mlfit(model, method, DE_population_size=DE_population_size, flux_estimation_MCMC=flux_estimation_MCMC,
                  fix_parameters_dictionnary=fix_parameters_dictionnary,
                  grid_resolution=grid_resolution, computational_pool=computational_pool, binary_regime=binary_regime)

        self.fits.append(fit)

    def telescopes_names(self):
        """Print the the telescope's names contain in the event.
        """
        print([self.telescopes[i].name for i in range(len(self.telescopes))])

    def check_event(self):
        """Function to check if everything is correctly set before the fit.
        An ERROR is returned if the check is not successful
        Should be used before any event_fit function calls

        First check if the event name is a string.
        Then check if the right ascension (event.ra) is between 0 and 360 degrees.
        Then check if the declination (event.dec) is between -90 and 90 degrees.
        Then check if you have any telescopes ingested.
        Finally check if your telescopes have a lightcurve attributes different from None.
        """

        if not isinstance(self.name, str):
            raise EventException('ERROR : The event name (' + str(
                self.name) + ') is not correct, it has to be a string')

        if (self.ra > 360) or (self.ra < 0):
            raise EventException('ERROR : The event ra (' + str(
                self.ra) + ') is not correct, it has to be a float between 0 and 360 degrees')

        if (self.dec > 90) or (self.dec < -90):
            raise EventException('ERROR : The event dec (' + str(
                self.dec) + ') is not correct, it has to be between -90 and 90 degrees')

        if len(self.telescopes) == 0:
            raise EventException('There is no telescope associated to your event, no fit possible!')

        else:

            for telescope in self.telescopes:

                if (len(telescope.lightcurve_magnitude) == 0) & \
                        (len(telescope.lightcurve_flux) == 0):
                    print('ERROR : There is no associated lightcurve in magnitude or flux with ' \
                          'this telescopes : ' \
                          + telescope.name + ', add one with telescope.lightcurve = your_data')
                    raise EventException('There is no lightcurve associated to the  telescope ' + str(
                        telescope.name) + ', no fit possible!')

        print(sys._getframe().f_code.co_name, ' : Everything looks fine, this event can be fitted')

    def find_survey(self, choice=None):
        """Function to find the survey telescope in the telescopes list,
        and put it on the first place (useful for some fits functions).

        :param choice: the name of the telescope choosing as the survey. Has to be a string.
                           Default is the first telescope.
        """
        self.survey = choice or self.telescopes[0].name

        names = [telescope.name for telescope in self.telescopes]
        if any(self.survey in name for name in names):

            index = np.where(self.survey == np.array(names))[0]
            sorting = np.arange(0, len(self.telescopes))
            sorting = np.delete(sorting, index)
            sorting = np.insert(sorting, 0, index)
            self.telescopes = [self.telescopes[i] for i in sorting]

        else:

            print('ERROR : There is no telescope names containing ' + self.survey)
            return

    def lightcurves_in_flux(self, choice='Yes'):
        """ Transform all telescopes magnitude lightcurves in flux units.


            :param choice: to clean your lightcurve or not. Has to be a string 'Yes' or 'No'.
            Defaul is 'Yes'. More details in the telescope module
        """

        for telescope in self.telescopes:
            telescope.lightcurve_flux = telescope.lightcurve_in_flux(choice)

    def compute_parallax_all_telescopes(self, parallax_model):
        """ Compute the parallax displacement for all the telescopes, if this is desired in
        the second order parameter.
        """

        for telescope in self.telescopes:

            if len(telescope.deltas_positions) == 0:
                telescope.compute_parallax(self, parallax_model)

    def total_number_of_data_points(self):
        """ Compute the parallax displacement for all the telescopes, if this is desired in
            the second order parameter.
            :return: n_data, the total number of points
            :rtype: float
        """
        n_data = 0.0

        for telescope in self.telescopes:
            n_data = n_data + telescope.n_data('flux')

        return n_data
